#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course api serialize module """
from rest_framework import serializers

from tesla_ce.models import Course
from ..vle import VLESerializer


class InstitutionCourseSerializer(serializers.ModelSerializer):
    """Course serialize class."""

    id = serializers.IntegerField()
    vle = VLESerializer(read_only=True)
    vle_course_id = serializers.CharField(read_only=True)
    code = serializers.CharField(read_only=True)
    description = serializers.CharField(read_only=True)
    start = serializers.DateTimeField(read_only=True)
    end = serializers.DateTimeField(read_only=True)
    user_roles = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = Course
        exclude = ["learners", "instructors"]

    def get_user_roles(self, instance):
        """
            Get the list of roles for a user in this course
            :param instance: Course instance
            :return: List of roles
        """
        roles = []
        if instance.learners.filter(id=self.context['request'].user.id).count() > 0:
            roles.append('LEARNER')
        if instance.instructors.filter(id=self.context['request'].user.id).count() > 0:
            roles.append('INSTRUCTOR')
        return roles



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution Instructor views module """
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionCourseInstructorSerializer
from tesla_ce.models import Instructor
from tesla_ce.models import Course


# pylint: disable=too-many-ancestors
class InstitutionCourseInstructorViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows instructors to be viewed or edited.
    """
    model = Instructor
    serializer_class = InstitutionCourseInstructorSerializer
    permission_classes = [
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminPermission |
        permissions.InstitutionLegalAdminReadOnlyPermission |
        permissions.InstitutionDataAdminReadOnlyPermission |
        permissions.InstitutionCourseLearnerReadOnlyPermission |
        permissions.InstitutionCourseInstructorPermission
    ]

    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['uid', 'first_name', 'last_name', 'email']
    search_fields = ['uid', 'first_name', 'last_name', 'email']


    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(Course.objects)
        queryset = queryset.get().instructors
        return queryset.all().order_by('id')



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Activity model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from tesla_ce.apps.api.utils import JSONField
from .base_model import BaseModel


class WebhookClient(BaseModel):
    """ Webhook clients model. """
    name = models.CharField(max_length=250, null=False, blank=False, unique=True,
                            help_text=_("Name that identifies the client."))

    client_header = models.CharField(max_length=250, null=False, blank=False, unique=True,
                                     help_text=_("Name of the request header used to identify the client."))

    id_header = models.CharField(max_length=250, null=True, blank=False, default=None,
                                 help_text=_("Name of the request header used to get the message id."))

    description = models.TextField(null=True, blank=True,
                                   help_text=_("Client description."))

    implementation = models.CharField(max_length=255, null=False, blank=True,
                                      help_text=_("Webhook implementation class."))

    enabled = models.BooleanField(null=None, blank=None, default=False,
                                  help_text=_("Whether this client is enabled or not"))

    credentials = models.TextField(null=False, blank=False,
                                   help_text=_("Client authentication credentials."))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<WebhookClient(id='%r', name='%r', enabled='%r')>" % (
            self.id, self.name, self.enabled)

    def get_credentials(self):
        """
            Returns the body of the message
            :return: The JSON body
            :rtype: dict
        """
        if self.body is not None:
            json_field = JSONField()
            return json_field.to_representation(self.body)

        return None

WEBHOOK_STATUS = (
    (0, _('Created')),
    (1, _('Processing')),
    (2, _('Processed')),
    (3, _('Error')),
    (4, _('Timeout')),
)


class WebhookMessage(BaseModel):
    """ Webhook message model. """
    client = models.ForeignKey(WebhookClient, null=False, on_delete=models.CASCADE)

    status = models.IntegerField(choices=WEBHOOK_STATUS, null=False,
                                 default=0,
                                 help_text=_('Status for this request'))
    message_id = models.CharField(max_length=250, null=False, blank=False,
                                  help_text=_("Message Id."))
    error = models.TextField(null=True, blank=True,
                             help_text=_("Error message."))
    body = models.TextField(null=False, blank=False,
                            help_text=_("Webhook message body."))
    start_process_at = models.DateTimeField(null=True, default=None,
                                            help_text=_('When the webhook started to be processed'))
    updated_at = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        unique_together = (('client', 'message_id'),)

    def get_body(self):
        """
            Returns the body of the message
            :return: The JSON body
            :rtype: dict
        """
        if self.body is not None:
            json_field = JSONField()
            return json_field.to_representation(self.body)

        return None



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""VLE api serialize module."""
from django.conf import settings
from rest_framework import serializers


class VLELauncherBodySerializer(serializers.Serializer):
    vle_user_uid = serializers.CharField(required=True, allow_blank=False, allow_null=False)
    target = serializers.ChoiceField(required=False, allow_null=False, default="DASHBOARD", choices=("DASHBOARD", "LAPI",))
    target_url = serializers.CharField(required=False, allow_null=True, default=None)
    ttl = serializers.IntegerField(required=False, allow_null=False, default=120)
    session_id = serializers.IntegerField(required=False, allow_null=True, default=None)

    class Meta:
        fields = ["vle_user_uid", "target", "target_url", "ttl", "session_id"]


class VLELauncherDataSerializer(serializers.Serializer):
    """New launcher data."""

    token = serializers.CharField(required=True, allow_blank=False, allow_null=False)
    id = serializers.IntegerField(required=True, allow_null=False)
    url = serializers.SerializerMethodField()

    class Meta:
        fields = ["id", "token", "url"]

    def get_url(self, object):
        return "{}/ui/auth/launcher?id={}&token={}".format(settings.DASHBOARD_URL, object['id'], object['token'])



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE deployment module """
import abc

from django.template.loader import render_to_string

from tesla_ce.lib.exception import TeslaDeploymentException
from ..modules import get_modules


class BaseDeployment:
    """ Base class for deployments """

    def __init__(self, client):
        """
            Default constructor

            :param client: Client instance
            :type client: Client
        """
        self._client = client
        self._config = client.config.config

    @classmethod
    def get_instance(cls, client, orchestrator=None):
        """
            Get specific deployment object for selected orchestrator

            :param client: TeSLA Client instance
            :type client: Client
            :param orchestrator: Orchestrator to be used
            :type orchestrator: str
        """
        if orchestrator is None:
            orchestrator = client.config.config.get('DEPLOYMENT_ORCHESTRATOR')
        instance = None
        if orchestrator == 'swarm':
            instance = SwarmDeployment(client)

        if instance is None:
            raise TeslaDeploymentException('Invalid orchestrator "{}"'.format(orchestrator))

        return instance

    @abc.abstractmethod
    def get_deployment_scripts(self):
        """
            Generate deployment scripts for provided class

            :return: All required scripts and files for the deployment
            :rtype: dict
        """
        pass

    @abc.abstractmethod
    def get_services_scripts(self):
        """
            Generate services deployment scripts for provided class

            :return: All required scripts and files for the deployment of required services
            :rtype: dict
        """
        pass

    @abc.abstractmethod
    def get_vle_scripts(self, vle):
        """
            Generate VLE deployment scripts for provided class

            :param vle: The VLE instance
            :type vle: tesla_ce.models.VLE
            :return: All required scripts and files for the deployment of required services
            :rtype: dict
        """
        pass

    @abc.abstractmethod
    def get_provider_scripts(self, provider, credentials=None):
        """
            Generate Provider deployment scripts for provided class

            :param provider: The Provider instance
            :type provider: tesla_ce.models.Provided
            :param credentials: List of credentials required by this provider
            :type credentials: list
            :return: All required scripts and files for the deployment of required services
            :rtype: dict
        """
        pass

    @abc.abstractmethod
    def get_status(self):
        """
            Get the deployment status

            :return: Deployment status
            :rtype: dict
        """
        pass

    @staticmethod
    def _remove_empty_lines(text):
        return '\n'.join([line for line in text.split('\n') if line.strip()])


class SwarmDeployment(BaseDeployment):

    def get_services_scripts(self):
        # Generate the context
        context = self._config.get_config_kv()

        # Generate services scripts
        files = {
            'tesla_services.yml': self._remove_empty_lines(render_to_string('deployment/swarm/tesla_services.yml',
                                                                            context)),
            'secrets/DB_ROOT_PASSWORD': context['DB_ROOT_PASSWORD'],
            'secrets/VAULT_DB_PASSWORD': context['VAULT_DB_PASSWORD'],
            'secrets/REDIS_PASSWORD': context['REDIS_PASSWORD'],
            'secrets/STORAGE_ACCESS_KEY': context['STORAGE_ACCESS_KEY'],
            'secrets/STORAGE_SECRET_KEY': context['STORAGE_SECRET_KEY'],
            'secrets/RABBITMQ_ADMIN_USER': context['RABBITMQ_ADMIN_USER'],
            'secrets/RABBITMQ_ADMIN_PASSWORD': context['RABBITMQ_ADMIN_PASSWORD'],
            'secrets/RABBITMQ_ERLANG_COOKIE': context['RABBITMQ_ERLANG_COOKIE'],
            'config/vault_config.json': self._remove_empty_lines(render_to_string('deployment/swarm/vault_config.json',
                                                                                  context)),
        }
        # Add loadbalancer script
        if self._config.get('DEPLOYMENT_LB') == 'traefik':
            files['tesla_lb.yml'] = self._remove_empty_lines(render_to_string('deployment/swarm/tesla_lb_traefik.yml',
                                                                              context))

        return files

    def get_deployment_scripts(self):
        # Generate the context
        context = self._config.get_config_kv()

        # Generate services scripts
        modules = get_modules()

        # Remove unnecessary workers
        if self._config.get('DEPLOYMENT_SPECIALIZED_WORKERS'):
            del modules['worker-all']
        else:
            del modules['worker-enrolment']
            del modules['worker-enrolment-storage']
            del modules['worker-enrolment-validation']
            del modules['worker-verification']
            del modules['worker-alerts']
            del modules['worker-reporting']

        context['services'] = [modules[module] for module in modules]
        files = {
            'tesla_core.yml': self._remove_empty_lines(render_to_string('deployment/swarm/tesla_core.yml',
                                                                        context)),
        }

        for module in modules:
            credentials = self._client.vault.get_module_credentials(module)
            files['secrets/{}_VAULT_ROLE_ID'.format(module.upper())] = credentials['role_id']
            files['secrets/{}_VAULT_SECRET_ID'.format(module.upper())] = credentials['secret_id']

        # Add loadbalancer script
        if self._config.get('DEPLOYMENT_LB') == 'traefik':
            files['tesla_lb.yml'] = self._remove_empty_lines(render_to_string('deployment/swarm/tesla_lb_traefik.yml',
                                                                              context))

        # Add Services script
        if self._config.get('DEPLOYMENT_SERVICES'):
            files['tesla_services.yml'] = self._remove_empty_lines(
                render_to_string('deployment/swarm/tesla_services.yml', context))

        # Add Front-end script
        files['tesla_dashboards.yml'] = self._remove_empty_lines(
            render_to_string('deployment/swarm/tesla_frontend.yml', context))

        return files

    def get_vle_scripts(self, vle):
        """
            Generate VLE deployment scripts for provided class

            :param vle: The VLE instance
            :type vle: tesla_ce.models.VLE
            :return: All required scripts and files for the deployment of required services
            :rtype: dict
        """
        # Generate the context
        context = self._config.get_config_kv()

        if context.get('MOODLE_ADMIN_EMAIL') is None:
            context['MOODLE_ADMIN_EMAIL'] = context['TESLA_ADMIN_MAIL']

        # Update VLE url
        vle.url = "https://moodle.{}".format(context['TESLA_DOMAIN'])
        vle.save()

        # Register the VLE
        vle_info = self._client.vault.register_vle(vle)

        # Generate vle scripts
        files = {
            'tesla_moodle.yml': self._remove_empty_lines(render_to_string('deployment/swarm/tesla_moodle.yml',
                                                                          context)),
            'secrets/MOODLE_DB_PASSWORD': context['MOODLE_DB_PASSWORD'],
            'secrets/MOODLE_ADMIN_PASSWORD': context['MOODLE_ADMIN_PASSWORD'],
            'secrets/MOODLE_ROLE_ID': vle_info['role_id'],
            'secrets/MOODLE_SECRET_ID': vle_info['secret_id'],
        }

        return files

    def get_provider_scripts(self, provider, credentials=None):
        """
            Generate Provider deployment scripts for provided class

            :param provider: The Provider instance
            :type provider: tesla_ce.models.Provided
            :param credentials: List of credentials required by this provider
            :type credentials: list
            :return: All required scripts and files for the deployment of required services
            :rtype: dict
        """
        # Generate the context
        context = self._config.get_config_kv()

        # Add provider base info
        context['provider'] = {
            'acronym': provider.acronym,
            'image': provider.image
        }

        # Add credentials
        if credentials is None:
            credentials = []
        # Add credentials
        context['credentials'] = []
        credential_values = {}
        for credential in credentials:
            cred_name = None
            cred_value = ""
            if isinstance(credential, dict):
                cred_name = credential['name']
                cred_value = credential['value']
            elif isinstance(credential, str):
                cred_name = credential
            # Store credentials
            if cred_name is not None:
                context['credentials'].append(cred_name)
                credential_values[cred_name] = cred_value

        # Get provider data
        provider_info = self._client.vault.register_provider(provider)

        # Generate provider scripts
        files = {
            'tesla_{}_provider.yml'.format(provider.acronym): self._remove_empty_lines(
                render_to_string('deployment/swarm/tesla_provider.yml', context)
            ),
            'secrets/{}_ROLE_ID'.format(provider.acronym.upper()): provider_info['role_id'],
            'secrets/{}_SECRET_ID'.format(provider.acronym.upper()): provider_info['secret_id'],
        }
        # Add credentials file
        for name in credential_values:
            files['secrets/{}_{}'.format(provider.acronym.upper(), name.upper())] = credential_values[name]

        return files

    def get_status(self):
        return {
            'status': 0,
            'warnings': 0,
            'errors': 0,
            'info': {}
        }



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Provider actions for Use Case tests """
import io
import mock
import random
import requests
import simplejson
import uuid

from django.utils import timezone

from tests import auth_utils
from tests.conftest import get_random_string

from ..worker import get_task_by_queue


def provider_validate_samples(providers, tasks):
    """
    Providers validate samples in their queues
    :param providers: Available providers
    :param tasks: Pending tasks to be processed
    :return: List of pending tasks
    """
    # List simulating the validation queue
    pending_tasks_validation = []

    def create_validation_summary(*args, **kwargs):
        pending_tasks_validation.append(('create_validation_summary', args, kwargs))

    # Split tasks in queues
    queues = get_task_by_queue(tasks)

    # Each provider should process their tasks
    for queue_name in queues:
        provider = None
        deferred = False
        # Get the provider with this queue
        for prov in providers:
            if providers[prov]['queue'] == queue_name:
                instrument_id = providers[prov]['instrument']['id']
                provider = providers[prov]['credentials']
                deferred = providers[prov]['deferred']
        assert provider is not None
        client, config = auth_utils.client_with_approle_credentials(provider['role_id'], provider['secret_id'])
        # Get the Provider ID from configuration
        provider_id = config['provider_id']

        task_count = 0
        for task in queues[queue_name]:
            # Get the parameters
            assert task[0] == 'validate_request'
            learner_id, sample_id, validation_id = task[1][0]
            learner_id = str(learner_id)

            # Get Sample information
            get_sample_resp = client.get('/api/v2/provider/{}/enrolment/{}/sample/{}/validation/{}/'.format(
                provider_id, learner_id, sample_id, validation_id)
            )
            assert get_sample_resp.status_code == 200
            sample = get_sample_resp.data

            # Download sample content
            sample_data_resp = requests.get(sample['sample']['data'], verify=False)
            assert sample_data_resp.status_code == 200
            sample['sample']['data'] = sample_data_resp.json()

            # Do validation
            val_result = {
                'status': (task_count % 2) + 1,  # Set half of the samples as not valid
                'error_message': None,
                'validation_info': {
                    'free_field1': 35,
                    'other_field':{
                        'status':3
                    }
                },
                'message_code_id': None,
                'contribution': 1.0 / (instrument_id * 2)  # We send four times samples than the instrument id
            }
            task_count +=1

            if deferred:
                # Deferred instrument
                samp_status_resp = client.post(
                    '/api/v2/provider/{}/enrolment/{}/sample/{}/validation/{}/status/'.format(
                        provider_id, learner_id, sample_id, validation_id),
                                              data={"status": 4},  # WAITING_EXTERNAL_SERVICE
                                              format='json'
                                              )
                assert samp_status_resp.status_code == 200
                def_validation_result = {
                    'status': 4,  # WAITING_EXTERNAL_SERVICE
                    'error_message': None,
                    'validation_info': None,
                    'message_code_id': None,
                    'contribution': None
                }
                def_result = {
                    'result': def_validation_result,
                    'status': 4,
                    'info': {
                        'info_field1': 56,
                        'info_other_field': 'test',
                        'test_data_for_notification': {
                            'type': 'validation',
                            'url': '/api/v2/provider/{}/enrolment/{}/sample/{}/validation/{}/'.format(
                                provider_id, learner_id, sample_id, validation_id
                            ),
                            'data': val_result
                        }
                    },
                    'learner_id': learner_id,
                    'sample_id': sample_id,
                    'validation_id': validation_id
                }
                # Set the result
                sample_deferred_validate_resp = client.put(def_result['info']['test_data_for_notification']['url'],
                                                           data=def_result['result'],
                                                           format='json')
                assert sample_deferred_validate_resp.status_code == 200

                # Send a notification request
                notification = {
                    'key': get_random_string(10),
                    'when': (timezone.now() + timezone.timedelta(seconds=-10)).isoformat(),  # Negative to ensure is due
                    'info': def_result['info']
                }
                send_notification_resp = client.post('/api/v2/provider/{}/notification/'.format(provider_id),
                                                     data=notification,
                                                     format='json')
                assert send_notification_resp.status_code == 201
            else:
                with mock.patch('tesla_ce.tasks.requests.enrolment.create_validation_summary.apply_async',
                                create_validation_summary):
                    send_validation_resp = client.put(
                        '/api/v2/provider/{}/enrolment/{}/sample/{}/validation/{}/'.format(
                            provider_id, learner_id, sample_id, validation_id),
                        data=val_result,
                        format='json'
                    )
                    assert send_validation_resp.status_code == 200

    return pending_tasks_validation


def provider_enrol_learners(providers, tasks):
    """
    Provider perform learners enrolment
    :param providers: Available providers
    :param tasks: Pending tasks to be processed
    :return: List of new pending tasks
    """
    # Split tasks in queues
    queues = get_task_by_queue(tasks)

    # Each provider should process their tasks
    for queue_name in queues:
        provider = None
        instrument_id = None
        deferred = False
        # Get the provider with this queue
        for prov in providers:
            if providers[prov]['queue'] == queue_name:
                instrument_id = providers[prov]['instrument']['id']
                provider = providers[prov]['credentials']
                deferred = providers[prov]['deferred']
        assert provider is not None
        assert instrument_id is not None
        client, config = auth_utils.client_with_approle_credentials(provider['role_id'], provider['secret_id'])
        # Get the Provider ID from configuration
        provider_id = config['provider_id']

        for task in queues[queue_name]:
            # Get the parameters
            assert task[0] == 'enrol_learner'
            learner_id, sample_id = task[1][0]
            learner_id = str(learner_id)
            task_id = uuid.uuid4()

            # Get the model
            get_model_resp = client.post(
                '/api/v2/provider/{}/enrolment/'.format(provider_id),
                data={
                     'learner_id': learner_id,
                     'task_id': str(task_id)
                },
                format='json'
            )
            assert get_model_resp.status_code == 201
            model = get_model_resp.data
            model_data = None
            if model is not None and model['model'] is not None:
                model_data_resp = requests.get(model['model'], verify=False)
                assert model_data_resp.status_code == 200
                model_data = model_data_resp.json()

            # Get validated samples
            final = False
            val_samples = []
            while not final:
                val_samples_resp = client.get('/api/v2/provider/{}/enrolment/{}/available_samples/'.format(
                    provider_id, learner_id))
                assert val_samples_resp.status_code == 200
                val_samples += val_samples_resp.data['results']
                if val_samples_resp.data['count'] <= len(val_samples):
                    final = True

            # If there is no sample, unlock the model
            if len(val_samples) == 0:
                unlock_resp = client.post('/api/v2/provider/{}/enrolment/{}/unlock/'.format(provider_id, learner_id),
                                          data={
                                              'token': task_id
                                          },
                                          format='json')
                assert unlock_resp.status_code == 200
                continue

            # Get sample validations
            for sample in val_samples:
                # Get available validations
                final = False
                sample['validations'] = []
                while not final:
                    sample_vals_resp = client.get('/api/v2/provider/{}/enrolment/{}/sample/{}/validation/'.format(
                        provider_id, learner_id, sample_id))
                    assert sample_vals_resp.status_code == 200
                    sample['validations'] += sample_vals_resp.data['results']
                    if sample_vals_resp.data['count'] <= len(sample['validations']):
                        final = True
                # Get the sample data
                sample_data_resp = requests.get(sample['data'], verify=False)
                assert sample_data_resp.status_code == 200
                sample['data'] = sample_data_resp.json()

            # Perform the enrolment
            if model_data is None:
                model_data = {}
            new_model_data = model_data
            if 'num_samples' not in new_model_data:
                new_model_data['num_samples'] = 0
            if 'instrument_id' not in new_model_data:
                new_model_data['instrument_id'] = instrument_id
            new_model_data['num_samples'] += len(val_samples)
            new_percentage = 0
            if 'percentage' in new_model_data:
                new_percentage = new_model_data['percentage']
            sample_list = []
            if 'used_samples' in model:
                sample_list = model['used_samples']
            for sample in val_samples:
                sample_list.append(sample['id'])
                for validation in sample['validations']:
                    new_percentage += validation['contribution']
            new_model_data['percentage'] = min(1.0, new_percentage)
            model['valid'] = True
            model['model'] = new_model_data
            model['can_analyse'] = True
            model['percentage'] = new_model_data['percentage']
            model['used_samples'] = sample_list

            enrolment_result = {
                'learner_id': learner_id,
                'task_id': task_id,
                'percentage': model['percentage'],
                'can_analyse': model['can_analyse'],
                'used_samples': model['used_samples'],
                'error_message': None
            }

            if deferred:
                def_enrolment_result = {
                    'status': 6,  # WAITING_EXTERNAL_SERVICE
                    'learner_id': learner_id,
                    'task_id': task_id
                }
                def_result = {
                    'result': def_enrolment_result,
                    'status': 6,
                    'model': model,
                    'info': {
                        'info_field1': 56,
                        'info_other_field': 'test',
                        'test_data_for_notification': {
                            'type': 'enrolment',
                            'url': '/api/v2/provider/{}/enrolment/{}/'.format(provider_id, learner_id),
                            'data': enrolment_result,
                            'model_data': {
                                'url': model['model_upload_url']['url'],
                                'data': model['model_upload_url']['fields'],
                                'file_content': model['model']
                            }
                        }
                    },
                    'learner_id': learner_id,
                    'sample_id': sample_id,
                    'task_id': task_id
                }
                # Set the result
                deferred_enrolment_resp = client.put(def_result['info']['test_data_for_notification']['url'],
                                                     data=def_result['result'],
                                                     format='json')
                assert deferred_enrolment_resp.status_code == 200

                # Send a notification request
                notification = {
                    'key': get_random_string(10),
                    'when': (timezone.now() + timezone.timedelta(seconds=-10)).isoformat(),  # Negative to ensure is due
                    'info': def_result['info']
                }
                send_notification_resp = client.post('/api/v2/provider/{}/notification/'.format(provider_id),
                                                     data=notification,
                                                     format='json')
                assert send_notification_resp.status_code == 201
            else:
                # Save the model data
                model_data_save_resp = requests.post(model['model_upload_url']['url'],
                                                     data=model['model_upload_url']['fields'],
                                                     files={
                                                         'file': io.StringIO(simplejson.dumps(model['model']))
                                                     }, verify=False)
                assert model_data_save_resp.status_code == 204

                # Save the model and unlock it
                model_save_resp = client.put('/api/v2/provider/{}/enrolment/{}/'.format(provider_id, learner_id),
                                             data={
                                                 'learner_id': learner_id,
                                                 'task_id': task_id,
                                                 'percentage': model['percentage'],
                                                 'can_analyse': model['can_analyse'],
                                                 'used_samples': model['used_samples']
                                             },
                                             format='json')
                assert model_save_resp.status_code == 200


def provider_verify_request(providers, tasks):
    """
        Providers verify the information from requests in their queues
        :param providers: Available providers
        :param tasks: Pending tasks to be processed
        :return: List of pending tasks
    """
    # List simulating the verification queue
    pending_tasks_verification = []

    def create_verification_summary_test(*args, **kwargs):
        pending_tasks_verification.append(('create_verification_summary', args, kwargs))

    # Split tasks in queues
    queues = get_task_by_queue(tasks)

    # Each provider should process their tasks
    for queue_name in queues:
        provider_creds = None
        instrument_id = None
        deferred = False
        # Get the provider with this queue
        for prov in providers:
            if providers[prov]['queue'] == queue_name:
                instrument_id = providers[prov]['instrument']['id']
                provider_creds = providers[prov]['credentials']
                deferred = providers[prov]['deferred']
        assert provider_creds is not None
        assert instrument_id is not None
        client, config = auth_utils.client_with_approle_credentials(provider_creds['role_id'],
                                                                    provider_creds['secret_id'])
        # Get the Provider ID from configuration
        provider_id = config['provider_id']

        for task in queues[queue_name]:
            # Get the parameters
            assert task[0] == 'verify_request'
            request_id, result_id = task[1][0]

            # Get the request
            get_request_resp = client.get('/api/v2/provider/{}/request/{}/'.format(provider_id, request_id))
            assert get_request_resp.status_code == 200
            request = get_request_resp.data
            learner_id = request['learner_id']

            # Get the provider data
            get_provider_resp = client.get('/api/v2/provider/{}/'.format(provider_id))
            assert get_provider_resp.status_code == 200
            provider = get_provider_resp.data

            # Get the model if necessary
            model_data = None
            if provider['instrument']['requires_enrolment']:
                # Download learner model
                get_model_resp = client.get('/api/v2/provider/{}/enrolment/{}/'.format(provider_id, learner_id))
                assert get_model_resp.status_code == 200
                model = get_model_resp.data
                assert model['can_analyse']

                # Get model data
                model_data_resp = requests.get(model['model'], verify=False)
                assert model_data_resp.status_code == 200
                model_data = model_data_resp.json()

            # Get request data
            request_data_resp = requests.get(request['request']['data'], verify=False)
            assert request_data_resp.status_code == 200
            request['request']['data'] = request_data_resp.json()

            # Instrument compute result without external services
            result_err = random.random() / 10.0  # Generate values with an error of maximum 10%
            if provider['inverted_polarity']:
                result = result_err
            else:
                result = 1.0 - result_err
            verification_result = {
                'status': 1,  # PROCESSED  (2 for ERROR)
                'error_message': None,
                'audit_data': {
                    'using_model': model_data is not None,
                    'some_other_field': get_random_string(15)
                },
                'result': result,
                'code': 1,  # OK  (2 WARNING, 3 ALERT)
                'message_code': None
            }

            # Perform verification
            if deferred:
                # Deferred instrument
                req_status_resp = client.post('/api/v2/provider/{}/request/{}/status/'.format(provider_id, request_id),
                                              data={"status": 7},  # WAITING_EXTERNAL_SERVICE
                                              format='json'
                                              )
                assert req_status_resp.status_code == 200
                def_verification_result = {
                    'status': 7,  # WAITING_EXTERNAL_SERVICE
                    'error_message': None,
                    'audit_data': {
                        'using_model': model_data is not None,
                        'some_other_field': get_random_string(15)
                    },
                    'result': None,
                    'code': 0,  # PENDING
                    'message_code': None
                }
                def_result = {
                    'result': def_verification_result,
                    'status': 7,
                    'info': {
                        'info_field1': 56,
                        'info_other_field': 'test',
                        'test_data_for_notification': {
                            'type': 'verification',
                            'url': '/api/v2/provider/{}/request/{}/'.format(provider_id, request_id),
                            'data': verification_result
                        }
                    },
                    'request_id': request_id,
                    'audit_data': {}
                }
                # Set the result
                req_deferred_verify_resp = client.put(def_result['info']['test_data_for_notification']['url'],
                                                      data=def_result['result'],
                                                      format='json')
                assert req_deferred_verify_resp.status_code == 200

                # Send a notification request
                notification = {
                    'key': get_random_string(10),
                    'when': (timezone.now() + timezone.timedelta(seconds=-10)).isoformat(),  # Negative to ensure is due
                    'info': def_result['info']
                }
                send_notification_resp = client.post('/api/v2/provider/{}/notification/'.format(provider_id),
                                                     data=notification,
                                                     format='json')
                assert send_notification_resp.status_code == 201
            else:
                with mock.patch('tesla_ce.tasks.requests.verification.create_verification_summary.apply_async',
                                create_verification_summary_test):
                    put_result_resp = client.put('/api/v2/provider/{}/request/{}/'.format(provider_id, request_id),
                                                 data=verification_result,
                                                 format='json'
                                             )
                    assert put_result_resp.status_code == 200

    return pending_tasks_verification


def provider_process_notifications(providers, tasks, expected_type):
    """
        Providers process the notifications in their queues
        :param providers: Available providers
        :param tasks: Pending tasks to be processed
        :param expected_type: Type of expected notifications
        :return: List of pending tasks
    """
    # List simulating the proper queue depending on type of notification
    pending_tasks = []

    def create_validation_summary_test(*args, **kwargs):
        pending_tasks.append(('create_validation_summary', args, kwargs))

    def create_verification_summary_test(*args, **kwargs):
        pending_tasks.append(('create_verification_summary', args, kwargs))

    # Split tasks in queues
    queues = get_task_by_queue(tasks)

    # Each provider should process their tasks
    for queue_name in queues:
        provider_creds = None
        instrument_id = None
        deferred = False
        # Get the provider with this queue
        for prov in providers:
            if providers[prov]['queue'] == queue_name:
                instrument_id = providers[prov]['instrument']['id']
                provider_creds = providers[prov]['credentials']
                deferred = providers[prov]['deferred']
        assert provider_creds is not None
        assert instrument_id is not None
        assert deferred
        client, config = auth_utils.client_with_approle_credentials(provider_creds['role_id'],
                                                                    provider_creds['secret_id'])
        # Get the Provider ID from configuration
        provider_id = config['provider_id']

        for task in queues[queue_name]:
            # Get the parameters
            assert task[0] == 'provider_notify'
            notification_id = task[1][0][0]

            # Get notification content
            get_notification_resp = client.get('/api/v2/provider/{}/notification/{}/'.format(provider_id,
                                                                                             notification_id))
            assert get_notification_resp.status_code == 200
            notification = get_notification_resp.data

            # Get the information required to process the notification
            notification_type = notification['info']['test_data_for_notification']['type']
            url = notification['info']['test_data_for_notification']['url']
            data = notification['info']['test_data_for_notification']['data']
            assert notification_type == expected_type
            # Process each request depending on their type
            if notification_type == 'validation':
                with mock.patch('tesla_ce.tasks.requests.enrolment.create_validation_summary.apply_async',
                                create_validation_summary_test):
                    put_result_resp = client.put(url,
                                                 data=data,
                                                 format='json'
                                                 )
                    assert put_result_resp.status_code == 200
            elif notification_type == 'enrolment':
                model_url = notification['info']['test_data_for_notification']['model_data']['url']
                model_data = notification['info']['test_data_for_notification']['model_data']['data']
                model_file_content = notification['info']['test_data_for_notification']['model_data']['file_content']

                # Save the model data
                model_data_save_resp = requests.post(model_url,
                                                     data=model_data,
                                                     files={
                                                         'file': io.StringIO(simplejson.dumps(model_file_content))
                                                     }, verify=False)
                assert model_data_save_resp.status_code == 204

                # Save the model and unlock it
                model_save_resp = client.put(url,
                                             data=data,
                                             format='json')
                assert model_save_resp.status_code == 200
            elif notification_type == 'verification':
                with mock.patch('tesla_ce.tasks.requests.verification.create_verification_summary.apply_async',
                                create_verification_summary_test):
                    put_result_resp = client.put(url,
                                                 data=data,
                                                 format='json'
                                                 )
                    assert put_result_resp.status_code == 200

            # Once processed, remove notification
            client.delete('/api/v2/provider/{}/notification/{}/'.format(provider_id, notification_id))

    return pending_tasks



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA Authentication serializers """
import json

from rest_framework import serializers

from tesla_ce.models import Institution
from tesla_ce.models import Provider
from tesla_ce.models import VLE
from tesla_ce.models.user import get_institution_roles
from tesla_ce.models.ui_option import get_user_ui_routes


class JSONField(serializers.JSONField):
    """ JSON field object used in serializers """

    def to_representation(self, value):
        """
            Parse input data
            :param value: The object value
            :return: A JSON representation
        """
        if isinstance(value, str) and len(value) == 0:
            value = None
        if value is not None and isinstance(value, str):
            try:
                return json.loads(value)
            except Exception:
                pass
        return super().to_representation(value)


class VLESerializer(serializers.ModelSerializer):
    """VLE serialize model module."""

    lti = JSONField()

    class Meta:
        model = VLE
        fields = "__all__"


class ProviderSerializer(serializers.ModelSerializer):
    """Provider serialize model module."""

    options = JSONField(allow_null=True)
    options_schema = JSONField(allow_null=True)

    class Meta:
        model = Provider
        fields = "__all__"


class AuthUserPasswordSerializer(serializers.Serializer):
    """ Body serializer for user/password authentication."""

    email = serializers.EmailField(required=True, allow_null=False)
    password = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    realm = serializers.CharField(allow_null=False, allow_blank=False, default="api")

    class Meta:
        fields = ["email", "password", "realm"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class AppRoleTokenSerializer(serializers.Serializer):
    """ Body serializer for approle based authentication."""

    role_id = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    secret_id = serializers.CharField(required=True, allow_null=False, allow_blank=False)

    class Meta:
        fields = ["role_id", 'secret_id']

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class AuthTokenSerializer(serializers.Serializer):
    """ Body serializer for token based authentication."""

    token = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    id = serializers.IntegerField(required=True, allow_null=False)

    class Meta:
        fields = ["token", "id"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class AuthTokenRefreshSerializer(serializers.Serializer):
    """ Body serializer for token refresh."""

    token = serializers.CharField(required=True, allow_null=False, allow_blank=False)

    class Meta:
        fields = ["token"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class AuthTokenPairSerializer(serializers.Serializer):
    """ Serializer for token pair."""

    access_token = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    refresh_token = serializers.CharField(required=True, allow_null=False, allow_blank=False)

    class Meta:
        fields = ["access_token", "refresh_token"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class AlternativeAuthTokenRefreshSerializer(serializers.Serializer):
    """ Body serializer for token refresh for Angular."""

    token = AuthTokenPairSerializer(required=True, allow_null=False)

    class Meta:
        fields = ["token"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class UserDataInstitutionSerializer(serializers.Serializer):
    """ Serializer for user data."""
    id = serializers.IntegerField(required=True, allow_null=False)
    acronym = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    name = serializers.CharField(required=True, allow_null=False, allow_blank=False)

    class Meta:
        fields = ["id", "acronym", "name"]

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')


class UserDataSerializer(serializers.Serializer):
    """ Serializer for user data."""
    id = serializers.IntegerField(required=True, allow_null=False)
    first_name = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    last_name = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    username = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    email = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    is_admin = serializers.BooleanField(source='is_staff', required=True, allow_null=False)
    full_name = serializers.SerializerMethodField()
    uid = serializers.CharField(required=False, allow_null=True, allow_blank=False, default=None)
    locale = serializers.CharField(required=False, allow_null=True, allow_blank=False, default=None)
    institution = serializers.SerializerMethodField(read_only=True)
    institutions = serializers.SerializerMethodField(read_only=True)
    roles = serializers.SerializerMethodField(read_only=True)
    routes = serializers.SerializerMethodField(read_only=True)

    class Meta:
        fields = ["first_name", "last_name", "username", "email", "is_admin", "full_name", "uid", "locale",
                  "institution", "roles"]

    @staticmethod
    def get_full_name(instance):
        """
            Get full name of the user
            :param instance: User instance
            :return: Full name
        """
        return instance.first_name + ' ' + instance.last_name

    @staticmethod
    def get_institution(instance):
        """
            Get institution information for user
            :param instance: User instance
            :return: Institution information
        """

        if hasattr(instance, "institutionuser"):
            roles = get_institution_roles(instance)
            try:
                learner_id = str(instance.institutionuser.learner.learner_id)
            except Exception:
                learner_id = None

            return {
                "id": instance.institutionuser.institution.id,
                "acronym": instance.institutionuser.institution.acronym,
                "uid": instance.institutionuser.uid,
                "roles": roles,
                "locale": instance.institutionuser.locale,
                "learner_id": learner_id
            }

        return None

    def get_institutions(self, instance):
        """
            Get list of institutions for user
            :param instance: User instance
            :return: Institutions list
        """
        institutions = []
        if 'GLOBAL_ADMIN' in self.get_roles(instance):
            for institution in Institution.objects.all().order_by('acronym'):
                institutions.append({
                    "id": institution.id,
                    "acronym": institution.acronym,
                    "uid": None,
                    "roles": ['GLOBAL_ADMIN'],
                    "locale": None,
                })
        else:
            default_institution = self.get_institution(instance)
            if default_institution is not None:
                institutions.append(default_institution)

        return institutions

    def get_roles(self, instance):
        """
            Get the list of roles for a user
            :param instance: User instance
            :return: List of roles
        """
        roles = []
        inst = self.get_institution(instance)
        if inst is not None:
            roles = inst['roles']
        if instance.is_staff:
            roles.append('GLOBAL_ADMIN')

        return roles

    @staticmethod
    def get_routes(instance):
        """
            Get the list of allowed routes for the user
            :param instance: User object
            :return: List of allowed routes
        """
        return get_user_ui_routes(instance)

    def update(self, instance, validated_data):
        """
            Update existing instance
            :param instance: The current instance
            :param validated_data: The new data
            :return: Updated instance
        """
        raise NotImplementedError('Operation not permitted')

    def create(self, validated_data):
        """
            Create a new instance
            :param validated_data: The data for the new instance
            :return: New instance
        """
        raise NotImplementedError('Operation not permitted')



# Generated by Django 3.0 on 2021-04-15 16:29

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0011_auto_20200930_1842'),
    ]

    operations = [
        migrations.AddField(
            model_name='institutionuser',
            name='data_admin',
            field=models.BooleanField(default=False, help_text='Whether this user can manage the data of the institution'),
        ),
        migrations.AlterField(
            model_name='enrolmentsamplevalidation',
            name='status',
            field=models.IntegerField(choices=[(0, 'Validating'), (1, 'Valid'), (2, 'Error'), (3, 'Timeout'), (4, 'Waiting External Service')], default=0, help_text='Validation status for this sample'),
        ),
        migrations.AlterField(
            model_name='requestproviderresult',
            name='status',
            field=models.IntegerField(choices=[(0, 'Pending'), (1, 'Processed'), (2, 'Error'), (3, 'Timeout'), (4, 'Missing Provider'), (5, 'Missing Enrolment'), (6, 'Processing'), (7, 'Waiting External Service')], default=0, help_text='Status for this result'),
        ),
        migrations.AlterField(
            model_name='requestresult',
            name='status',
            field=models.IntegerField(choices=[(0, 'Pending'), (1, 'Processed'), (2, 'Error'), (3, 'Timeout'), (4, 'Missing Provider'), (5, 'Missing Enrolment'), (6, 'Processing'), (7, 'Waiting External Service')], default=0, help_text='Status for this result'),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Public Bucket definition module """
from django.conf import settings
from storages.backends.s3boto3 import S3Boto3Storage


class PublicS3Boto3Storage(S3Boto3Storage):
    """
        Storage Bucket for public data (static files)
    """
    bucket_acl = 'download'
    default_acl = 'download'
    bucket_name = settings.AWS_S3_STORAGE_PUBLIC_BUCKET_NAME
    auto_create_bucket = True

    def __init__(self, acl=None, bucket=None, **settings):
        if acl is not None:
            self.bucket_acl = acl
        if bucket is not None:
            self.bucket_name = bucket
        super().__init__(**settings)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Providers notifications tasks module """
from django.utils import timezone

from tesla_ce import celery_app
from tesla_ce import models


@celery_app.task(ignore_result=True)
def send_provider_notifications():
    """
        Send notifications to providers
    """
    for notification in models.ProviderNotification.objects.filter(when__lte=timezone.now()):
        provider_notify.apply_async((notification.id,), queue=notification.provider.queue)


@celery_app.task(ignore_result=True)
def provider_notify(notification_id):
    """
        Process a provider notification
        :param notification_id: Notification ID
    """
    # Executed on provider
    pass



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution management """
import logging

import pytest

import tests.utils

from tests.utils import getting_variables


@pytest.mark.django_db
def test_api_institution_institution(rest_api_client, user_global_admin, institution_course_test_case):
    """

    INSTITUTION TEST: INSTITUTION
       1) List institutions
       2) Read institution information
       3) Change institution properties
    """
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id

    # Check that unauthenticated calls are not allowed
    user_global_admin.is_staff = False
    user_global_admin.save()
    # TODO: add list of not allowed users as 3rd parameter to check_403 function
    tests.utils.check_403(rest_api_client, '/api/v2/institution/', [user_global_admin])

    # 666 Authorization – JWT with Global Administration privileges
    user_global_admin.is_staff = True
    user_global_admin.save()

    # Set global admin user.
    rest_api_client.force_authenticate(user=user_global_admin)

    logging.info('\nTESTING INSTITUTIONS: Institutions *********************************************')

    # Get the list of Institutions
    """ ---------------------------------------------------------------------
    LIST INSTITUTIONS:
       GET /api/v2/institution/
       Status Codes:
           200 OK – Ok
    Request Headers: Authorization - JWT with Global Administration privileges
    """
    logging.info('\n1) LIST INSTITUTIONS --------------------------------------')
    body = tests.utils.get_rest_api_client(rest_api_client, '/api/v2/institution/',
                                           'List institutions', 'RESPONSE:', 200)
    variables = getting_variables(body, institution_id)
    if variables['n_institution'] > 0:
        tests.utils.check_pagination(rest_api_client, body)
        args = [['List Institutions'], ['ID FIRST INSTITUTION: ', variables['id_first_institution']],
                ['NUMBER OF INSTITUTIONS:', variables['n_institution']],
                ['INSTITUTION ID TEST', institution_id],
                ['ID NON EXISTING INSTITUTION:', variables['id_non_existing_institution']]]
    else:
        args = [['List Institutions'], ['Empty Institutions list. Reading first Institution test will be skipped.']]

    tests.utils.print_log(args)


    """ ---------------------------------------------------------------------
    READ INSTITUTION INFORMATION:
       GET /api/v2/institution/(int:institution_id)/
       Status Codes:
           200 OK – Ok
           404 Not Found - Institution not found

    Request Headers: Authorization - JWT with Institution privileges
    """
    # Read 1st institution information
    logging.info('\n2) READ 1st INSTITUTION INFORMATION --------------------------------------')
    # 666 TODO: READ institutions to DOC
    rest_api_client.force_authenticate(user=institution_user)

    if not variables['institution_empty']:
        str_url = '/api/v2/institution/' + str(institution_id) + '/'
        str_response = 'RESPONSE id=' + str(institution_id) + ':'
        tests.utils.get_rest_api_client(rest_api_client,
                                        str_url,
                                        'Read Institution Information', str_response, 200)

    # TODO? READ non existing instrument with numeric id - Status 404

    """ ---------------------------------------------------------------------
    CHANGE INSTITUTION PROPERTIES:
       PUT /api/v2/institution/(int:institution_id)/
       Status Codes:
           200 OK – Ok
           400 Bad Request - Invalid information provided. The response contains the description of the errors.
           404 Not Found – Institution not found

    Request Headers: Authorization - JWT with Institution Administration privileges
    """
    # [Status: 200] UPDATE (Change) institution properties
    logging.info('\n3) CHANGE INSTITUTION PROPERTIES --------------------------------------')

    # Privileges: Institution Admin
    institution_user.inst_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    # 666     if not institution_empty:
    str_data = {'acronym': 'CHANGED', 'name': 'CHANGED_NAME', 'external_ic': True,
                'mail_domain': 'CHANGED_MAIL_DOMAIN.COM', "disable_vle_learner_creation": True,
                'disable_vle_instructor_creation': True, 'disable_vle_user_creation': True}
    str_path = '/api/v2/institution/' + str(institution_id) + '/'
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Change Institution Properties', 'RESPONSE:', 200)

    # [Status: 400] Invalid information provided. The response contains the description of the errors.
    str_data = {'external_ic': 23}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Change Institution Properties', 'RESPONSE:', 400)

    # TODO? [Status: 404] UPDATE NON EXISTING instrument



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Monitoring status model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel

RESOURCE_TYPE = (
    (0, _('Custom')),
    (1, _('Queue')),
    (2, _('Service')),
    (3, _('Provider')),
    (4, _('Instrument')),
)

RESOURCE_STATUS = (
    (0, _('Unknown')),
    (1, _('Ready')),
    (2, _('Error')),
    (3, _('Warning')),
)


class Monitor(BaseModel):
    """ Monitor status model. """
    status = models.IntegerField(choices=RESOURCE_STATUS, null=False,
                                 default=0,
                                 help_text=_('Status of the resource'))

    type = models.IntegerField(choices=RESOURCE_TYPE, null=False,
                               default=0,
                               help_text=_('Type of the resource'))

    key = models.CharField(max_length=15, null=True, default=None, blank=False,
                           help_text=_('key of the resource'))

    info = models.TextField(null=False, blank=False, help_text=_("Resource information."))

    created_at = models.DateTimeField(auto_now_add=True)

    def __repr__(self):
        return "<Monitor(status='%r', type='%r', key='%r')>" % (
            self.get_status_display(), self.get_type_display(), self.key
        )



#  Copyright (c) 2021 Mireia Bellot
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for institution SEND data management """
import logging

import pytest

import tests.utils


@pytest.mark.django_db
def test_api_institution_send(rest_api_client, institution_course_test_case):
    institution_user = institution_course_test_case['user'].institutionuser
    institution_id = institution_course_test_case['institution'].id

    # List SEND Categories
    logging.info('\n1) List SEND Categories --------------------------------------')
    # Set Institution Admin/SEND privileges
    institution_user.send_admin = True
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/send/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List SEND categories', str_message, 200)
    n_send = body['count']
    assert n_send == 0

    # Create a new SEND Category
    logging.info('\n2) Create a new SEND Category --------------------------------------')
    data = {'enabled_options': ['text_to_speech', 'big_fonts', 'high_contrast'],
            'disabled_instruments': [1, 2, 3, 4, 5]}

    str_data = {'parent': '',
                'description': 'SEND Category for TESTING purposes',
                'data': data}
    new_send_id = tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                                   'Create a new SEND Category', 'RESPONSE: ', 201)

    # Ensure number of categories has increased
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List SEND categories', str_message, 200)
    assert n_send + 1 == body['count']

    # Read SEND Category information
    logging.info('\n3) Read SEND Category information --------------------------------------')

    str_path = '/api/v2/institution/{}/send/{}/'.format(institution_id, new_send_id)
    str_response = 'RESPONSE SEND ID={}:'.format(new_send_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'Read SEND Category information', str_response, 200)

    # Update SEND Category
    logging.info('\n4) Update SEND Category --------------------------------------')

    str_data = {'description': 'CHANGED SEND Category for TESTING purposes'}
    tests.utils.put_rest_api_client(rest_api_client, str_path, str_data,
                                    'Update SEND Category', str_response, 200)

    # Delete SEND Category
    logging.info('\n5) Delete SEND Category --------------------------------------')

    str_path = '/api/v2/institution/{}/send/{}/'.format(institution_id, new_send_id)
    tests.utils.delete_rest_api_client(rest_api_client, str_path,
                                       'Delete SEND Category', "RESPONSE: ", 204)

    # Ensure number of SEND categories has decreased
    str_path = '/api/v2/institution/{}/send/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    body = tests.utils.get_rest_api_client(rest_api_client, str_path,
                                           'List SEND categories', str_message, 200)
    assert body['count'] == n_send

    # Ensure SEND categories cannot be listed with normal user
    # Remove Institution SEND privileges
    institution_user.inst_admin = False
    institution_user.send_admin = False
    institution_user.save()
    rest_api_client.force_authenticate(user=institution_user)

    str_path = '/api/v2/institution/{}/send/'.format(institution_id)
    str_message = 'RESPONSE institution_id:{}'.format(institution_id)
    tests.utils.get_rest_api_client(rest_api_client, str_path,
                                    'List SEND categories with invalid user', str_message, 403)

    # Create a new SEND Category with invalid user
    logging.info('\n2) Create a new SEND Category with invalid user  --------------------------------------')
    str_data = {'description': 'SEND Category for TESTING purposes',
                'data': ''}
    tests.utils.post_rest_api_client(rest_api_client, str_path, str_data,
                                     'Create a new SEND Category with invalid user', 'RESPONSE: ', 403)



# Generated by Django 3.0.8 on 2020-07-21 08:22

from django.db import migrations
from django.db import models


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0007_auto_20200712_1330'),
    ]

    operations = [
        migrations.AlterField(
            model_name='requestproviderresult',
            name='result',
            field=models.FloatField(default=0.0, help_text='Normalized result value', null=True),
        ),
        migrations.AlterField(
            model_name='requestresult',
            name='result',
            field=models.FloatField(default=0.0, help_text='Normalized result value, summarizing results from providers', null=True),
        ),
    ]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Vault setup module """
import random
import string
import time

import hvac
from hvac.exceptions import InvalidPath
from hvac.exceptions import InvalidRequest

from .policies import get_policies
from .policies import get_service_config_map
from ..config.conf import Config
from ..modules import get_modules
from ..modules import get_provider_properties
from ..modules import get_vle_properties

# Template to generate module JWT key names
MODULE_KEY_STR_TEMPLATE = 'jwt_module_{}'


class VaultSetup:
    """ Class that defines the methods to setup Vault"""

    #: Vault client
    _client: hvac.Client = None

    #: Configuration object
    _config: Config = None

    def __init__(self, client, config):
        """
            Default constructor

            :param client: Vault client
            :type client: hvac.Client
            :param config: Configuration object
            :type config tesla-lib.config.conf.Config

        """
        self._client = client
        self._config = config

        # Read Vault mount points
        self._approle_mount_point = self._config.get('VAULT_MOUNT_PATH_APPROLE')
        self._transit_mount_point = self._config.get('VAULT_MOUNT_PATH_TRANSIT')
        self._kv_mount_point = self._config.get('VAULT_MOUNT_PATH_KV')
        self._policy_prefix = self._config.get('VAULT_POLICIES_PREFIX')

    def run_setup(self):
        """
            Perform Vault setup
        """
        # Setup KV
        self.setup_kv()

        # Update configuration
        self._update_configuration()

        # Setup JWT infrastructure
        self.setup_jwt()

        # Setup policies
        self.setup_policies()

        # Setup authentication roles
        self.setup_roles()

    def setup_kv(self):
        """
            Setup the Key-Value secrets engine
        """
        from tesla_ce.client import get_version

        # Enable KV engine
        if not self._is_secret_engine_enabled('{}/'.format(self._kv_mount_point)):
            self._client.sys.enable_secrets_engine('kv', path=self._kv_mount_point, options={'version': 2},
                                                   description='TeSLA CE Secrets Engine')
            # This can take a time to be ready. Wait until active
            time.sleep(2)
            ready = False
            max_try = 5
            while not ready and max_try > 0:
                try:
                    self._client.secrets.kv.v2.create_or_update_secret(
                        path='{}/{}'.format('system', 'version'),
                        secret=dict({
                            'tesla-ce': get_version()
                        }),
                        mount_point=self._kv_mount_point
                    )
                    ready = True
                except (InvalidRequest, InvalidRequest):
                    time.sleep(5)
                    max_try -= 1

        # Update system version and vault mount paths
        self._client.secrets.kv.v2.create_or_update_secret(
            path='{}/{}'.format('system', 'version'),
            secret=dict({
                'tesla-ce': get_version()
            }),
            mount_point=self._kv_mount_point
        )

        # Check secret keys
        if self._config.get('DJANGO_SECRET_KEY') is None:
            # Generate a new secret key
            key = ''.join([string.ascii_letters,
                           string.digits,
                           string.punctuation]).replace('\'', '').replace('"', '').replace('\\', '')
            key = ''.join([random.SystemRandom().choice(key) for _ in range(50)])
            self._config.set('DJANGO_SECRET_KEY', key)

        # Store configuration values
        config = self._config.get_config()
        for section in config:
            for key in config[section]:
                self._client.secrets.kv.v2.create_or_update_secret(
                    path='config/{}/{}'.format(section, key),
                    secret=dict(config[section][key]),
                    mount_point=self._kv_mount_point
                )

    def _adapt_policies_path(self, policies):
        """
            Move policies to the defined Vault mount paths
            :param policies: Dictionary of policies with standard paths
            :return: Dictionary containing the policies with modified paths
        """
        adapted_policies = {}
        for policy in policies:
            adapted_policies[policy] = {'path': {}}
            for key in policies[policy]['path']:
                adapted_path = key.replace(
                    '<kv_path>', self._kv_mount_point
                ).replace(
                    '<transit_path>', self._transit_mount_point
                ).replace(
                    '<approle_path>', self._approle_mount_point
                ).replace(
                    '<policy_prefix>', self._policy_prefix
                )
                adapted_policies[policy]['path'][adapted_path] = policies[policy]['path'][key]
        return adapted_policies

    def setup_policies(self):
        """
            Creates the policies to grant access to services
        """
        # Create generic policies
        policies = self._adapt_policies_path(get_policies())
        for policy in policies:
            self._client.sys.create_or_update_policy(
                name='{}{}'.format(self._policy_prefix, policy),
                policy=policies[policy],
            )

    def _create_jwt_key(self, key):
        """
            Create an encryption key if it does not exist

            :param key: Name of the key
            :type key: str
        """
        # Create the key if it does not exist
        try:
            self._client.secrets.transit.read_key(key, mount_point=self._transit_mount_point)
        except InvalidPath:
            self._client.secrets.transit.create_key(key, key_type='rsa-4096', mount_point=self._transit_mount_point)

    def setup_jwt(self):
        """
            Setup JWT infrastructure
        """
        # Enable transit secret backend
        if not self._is_secret_engine_enabled('{}/'.format(self._transit_mount_point)):
            self._client.sys.enable_secrets_engine('transit', path=self._transit_mount_point)

        # Create a default key if it does not exist
        self._create_jwt_key('jwt_default')

        # Create learners key if it does not exist
        self._create_jwt_key('jwt_learners')

        # Create instructors key if it does not exist
        self._create_jwt_key('jwt_instructors')

        # Create users key if it does not exist
        self._create_jwt_key('jwt_users')

        # Create modules key if it does not exist
        self._create_jwt_key('jwt_modules')

        # Create individual modules key if they do not exist
        for module in get_modules():
            self._create_jwt_key(MODULE_KEY_STR_TEMPLATE.format(module))

    def setup_roles(self):
        """
            Creates the roles for the different modules that needs to authenticate with Vault
        """
        if not self._is_auth_method_enabled('{}/'.format(self._approle_mount_point)):
            self._client.sys.enable_auth_method('approle', path=self._approle_mount_point,
                                                description='TeSLA CE modules authentication')

        modules = get_modules()
        for module in modules:
            self._create_module_entity(modules[module])

    def _is_secret_engine_enabled(self, mount):
        """
            Check if a Vault secret engine is enabled

            :param mount: Mount point of the secret engine
            :return: True if the secret engine is enabled or False otherwise
            :rtype: bool
        """
        secrets_engines_list = self._client.sys.list_mounted_secrets_engines()['data']

        for engine in secrets_engines_list:
            if engine == mount:
                return True
        return False

    def _is_auth_method_enabled(self, mount):
        """
            Check if a Vault authentication method is enabled

            :param mount: Mount point of the authentication method
            :return: True if the authentication method is enabled or False otherwise
            :rtype: bool
        """
        auth_methods_list = self._client.sys.list_auth_methods()['data']

        for method in auth_methods_list:
            if method == mount:
                return True
        return False

    @staticmethod
    def _add_generic_service_access(module, conf_map, config, policies):
        """
            Update initial list of policies and configuration values with the ones required for services
            :param module: Module information
            :param conf_map: Configuration map for the module
            :param config: Initial set of configuration values
            :param policies: Initial set of policies
            :return: Updated configuration and policies
        """
        for service in module['services']:
            roles = module['services'][service]
            if service in conf_map:
                for role in roles:
                    map_val = conf_map
                    for k in role.split('/'):
                        if k not in map_val:
                            map_val = None
                            break
                        map_val = map_val[k]
                    if map_val is not None:
                        for value in map_val['map']:
                            config.add(value)
                        policies.add(map_val['policy'])

        return config, policies

    def _build_config_and_policies(self, module):
        """
            Build the set of configurations and policies for a module
            :param module: Module information
            :return: Configuration set and policies
        """

        # Build configuration map and required policies
        policies = set()
        config = set()
        conf_map = get_service_config_map()

        # Add access to public configuration
        policies.add(conf_map['public']['policy'])
        for value in conf_map['public']['map']:
            config.add(value)

        # Add access to DJango configuration
        if 'django' in module['services']:
            policies.add('django_config')

        # Add default modules policy
        policies.add('modules_{}'.format(module['module']))

        # Add generic services access
        config, policies = self._add_generic_service_access(module, conf_map, config, policies)

        # Prepend policy prefix to policies
        policies = ['{}{}'.format(self._policy_prefix, policy) for policy in policies]

        return config, policies

    def _create_module_policy(self, module_name):
        """
            Create the default ACL for module
            :param module_name: Module name
        """

        module_policy = {
                'path': {
                    '{}/data/modules/{}/*'.format(self._kv_mount_point, module_name): {
                        'capabilities': ['read', 'list']
                    },
                    '{}/verify/jwt_module_{}'.format(self._transit_mount_point, module_name): {
                        'capabilities': ['create', 'update']
                    },
                    '{}/verify/jwt_module_{}/*'.format(self._transit_mount_point, module_name): {
                        'capabilities': ['create', 'update']
                    }
                }
            }

        self._client.sys.create_or_update_policy(
            name='{}modules_{}'.format(self._policy_prefix, module_name),
            policy=module_policy,
        )

    def _create_module_entity(self, module, extra_data=None, module_name=None):
        """
            Create or update Vault entity and policies for a given module

            :param module: Module description tuple
            :type module: dict
            :param extra_data: Additional arguments to add to entity metadata
            :type extra_data: dict
            :param module_name: Name for the new module. If not provided the name in the module is used.
            :type module_name: str
        """

        # Set the module name
        if module_name is None:
            module_name = module['module']

        # Create the module policy
        self._create_module_policy(module_name)

        # Build configuration map and required policies
        config, policies = self._build_config_and_policies(module)

        # Create the metadata
        entity_metadata = {
            'module': module_name,
            'description': module['description'],
            'config': 'modules/{}/config'.format(module_name),
            'apps': 'modules/{}/apps'.format(module_name),
            'services': 'modules/{}/services'.format(module_name),
            'domain': '{}.{}'.format(module_name, self._config.get('TESLA_DOMAIN')),
        }

        # Add queues in case they are available
        if 'queues' in module['deployment']:
            entity_metadata['queues'] = ','.join(module['deployment']['queues'])

        # Add extra data
        if extra_data is not None:
            for data_key in extra_data:
                entity_metadata[data_key] = extra_data[data_key]

        # Write entity configuration keys
        self._client.secrets.kv.v2.create_or_update_secret(
            path='modules/{}/metadata'.format(module_name),
            secret={'data': entity_metadata},
            mount_point=self._kv_mount_point
        )
        self._client.secrets.kv.v2.create_or_update_secret(
            path='modules/{}/config'.format(module_name),
            secret={'data': list(config)},
            mount_point=self._kv_mount_point
        )
        self._client.secrets.kv.v2.create_or_update_secret(
            path='modules/{}/apps'.format(module_name),
            secret={'data': module['apps']},
            mount_point=self._kv_mount_point
        )
        self._client.secrets.kv.v2.create_or_update_secret(
            path='modules/{}/services'.format(module_name),
            secret={'data': module['services']},
            mount_point=self._kv_mount_point
        )

        # Create a role for this module
        self._client.auth.approle.create_or_update_approle(
            module_name,
            mount_point=self._approle_mount_point,
            token_policies=list(policies)
        )

        # Create Celery credentials for all modules
        # TODO: Use per module credentials
        if 'celery' in module['services']:
            prefix = module_name
            if module['module'].startswith('provider'):
                prefix = 'provider'
            self._client.secrets.kv.v2.create_or_update_secret(
                path='config/celery/credentials/{}_user'.format(prefix),
                secret={
                    'description': 'User to authenticate with broker',
                    'value': self._config.get('CELERY_BROKER_USER')
                },
                mount_point=self._kv_mount_point
            )
            self._client.secrets.kv.v2.create_or_update_secret(
                path='config/celery/credentials/{}_password'.format(prefix),
                secret={
                    'description': 'Password to authenticate with broker',
                    'value': self._config.get('CELERY_BROKER_PASSWORD')
                },
                mount_point=self._kv_mount_point
            )

    def _update_configuration(self):
        """
            Update general configuration on KV store
        """
        # Get current configuration
        try:
            tesla_config = self._client.secrets.kv.v2.read_secret_version(path='config/public/config',
                                                                          mount_point=self._kv_mount_point
                                                                          )['data']['data']
        except InvalidPath:
            tesla_config = {}

        # Update with new configuration values
        tesla_config['TESLA_DOMAIN'] = self._config.get('TESLA_DOMAIN')

        self._client.secrets.kv.v2.create_or_update_secret(
            path='config/public/config',
            secret=dict(tesla_config),
            mount_point=self._kv_mount_point
        )

    def register_vle(self, vle):
        """
            Register a new VLE on vault or update current information

            :param vle: VLE model instance
            :type vle: tesla_ce.models.VLE
        """
        vle_template = get_vle_properties()
        module_name = 'vle_{}'.format(str(vle.id).zfill(3))
        vle_template['module'] = module_name
        extra_data = {
            'vle_name': vle.name,
            'vle_id': vle.id,
            'vle_url': vle.url,
            'institution_acronym': vle.institution.acronym,
            'institution_id': vle.institution.id
        }

        self._create_module_entity(vle_template, extra_data, module_name)

        self._create_jwt_key(MODULE_KEY_STR_TEMPLATE.format(module_name))

        return module_name

    def register_provider(self, provider):
        """
            Register a new Provider on vault or update current information

            :param provider: Provider model instance
            :type provider: tesla_ce.models.Provider
        """
        provider_template = get_provider_properties()
        module_name = 'provider_{}'.format(str(provider.id).zfill(3))
        provider_template['module'] = module_name
        extra_data = {
            'provider_name': provider.name,
            'provider_id': provider.id,
            'provider_acronym': provider.acronym,
            'provider_queue': provider.queue,
            'instrument_id': provider.instrument.id
        }

        self._create_module_entity(provider_template, extra_data, module_name)

        self._create_jwt_key(MODULE_KEY_STR_TEMPLATE.format(module_name))

        return module_name



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Policies description module """


def get_service_config_map():
    """
        Get the configuration mappings for all roles
        :return: Configuration and credentials for the different services and roles
        :rtype: dict
    """
    return {
        'public': {
            'map': [
                ('TESLA_DOMAIN', 'config/tesla/domain'),
            ],
            'policy': 'public'
        },
        'db': {
            'config': {
                'map': [
                    ('DB_HOST', 'config/db/host'),
                    ('DB_PORT', 'config/db/port'),
                    ('DB_ENGINE', 'config/db/engine'),
                    ('DB_NAME', 'config/db/name'),
                ],
                'policy': 'db_config'
            },
            'credentials': {
                'admin': {
                    'map': [
                        ('DB_USER', 'config/db/root_user'),
                        ('DB_PASSWORD', 'config/db/root_password'),
                    ],
                    'policy': 'db_credentials_admin',
                },
                'rw_access': {
                    'map': [
                        ('DB_USER', 'config/db/user'),
                        ('DB_PASSWORD', 'config/db/password'),
                    ],
                    'policy': 'db_credentials_rw_access',
                }
            }
        },
        'vault': {
            'vle': {
                'write': {
                    'map': [],
                    'policy': 'vault_vle_write'
                }
            },
            'provider': {
                'write': {
                    'map': [],
                    'policy': 'vault_provider_write'
                }
            },
            'policy': {
                'write': {
                    'map': [],
                    'policy': 'vault_policy_write'
                }
            },
            'jwt': {
                'default': {
                    'create': {
                        'map': [],
                        'policy': 'vault_jwt_default_create'
                    },
                    'validate': {
                        'map': [],
                        'policy': 'vault_jwt_default_validate'
                    }
                },
                'users': {
                    'create': {
                        'map': [],
                        'policy': 'vault_jwt_users_create'
                    },
                    'validate': {
                        'map': [],
                        'policy': 'vault_jwt_users_validate'
                    }
                },
                'learners': {
                    'create': {
                        'map': [],
                        'policy': 'vault_jwt_learners_create'
                    },
                    'validate': {
                        'map': [],
                        'policy': 'vault_jwt_learners_validate'
                    }
                },
                'instructors': {
                    'create': {
                        'map': [],
                        'policy': 'vault_jwt_instructors_create'
                    },
                    'validate': {
                        'map': [],
                        'policy': 'vault_jwt_instructors_validate'
                    }
                },
                'modules': {
                    'create': {
                        'map': [],
                        'policy': 'vault_jwt_modules_create'
                    },
                    'validate': {
                        'map': [],
                        'policy': 'vault_jwt_modules_validate'
                    }
                },
                'module': {
                    'all': {
                        'validate': {
                            'map': [],
                            'policy': 'vault_jwt_module_all_validate'
                        },
                        'create': {
                            'map': [],
                            'policy': 'vault_jwt_module_all_create'
                        }
                    }
                }
            }
        },
        'django': {
            'config': {
                'map': [
                    ('DJANGO_SECRET_KEY', 'config/django/secret_key'),
                    ('DJANGO_ALLOWED_HOSTS', 'config/django/allowed_hosts'),
                ],
                'policy': 'django_config'
            }
        },
        'redis': {
            'config': {
                'map': [
                    ('REDIS_HOST', 'config/redis/host'),
                    ('REDIS_PORT', 'config/redis/port'),
                    ('REDIS_PASSWORD', 'config/redis/password'),
                    ('REDIS_DATABASE', 'config/redis/database'),
                ],
                'policy': 'redis_config'
            }
        },
        'storage': {
            'config': {
                'map': [
                    ('STORAGE_URL', 'config/storage/url'),
                    ('STORAGE_BUCKET_NAME', 'config/storage/bucket_name'),
                    ('STORAGE_REGION', 'config/storage/region'),
                    ('STORAGE_ACCESS_KEY', 'config/storage/access_key'),
                    ('STORAGE_SECRET_KEY', 'config/storage/secret_key'),
                ],
                'policy': 'storage_config'
            }
        },
        'rabbitmq': {
            'config': {
                'map': [
                    ('RABBITMQ_ADMIN_USER', 'config/rabbitmq/admin_user'),
                    ('RABBITMQ_ADMIN_PASSWORD', 'config/rabbitmq/admin_password'),
                    ('RABBITMQ_ERLANG_COOKIE', 'config/rabbitmq/erlang_cookie'),
                    ('RABBITMQ_PORT', 'config/rabbitmq/port'),
                    ('RABBITMQ_ADMIN_PORT', 'config/rabbitmq/admin_port'),
                ],
                'policy': 'rabbitmq_config'
            }
        },
        'celery': {
            'config': {
                'map': [
                    ('CELERY_BROKER_PROTOCOL', 'config/celery/broker_protocol'),
                    ('CELERY_BROKER_HOST', 'config/celery/broker_host'),
                    ('CELERY_BROKER_PORT', 'config/celery/broker_port'),
                    ('CELERY_BROKER_REGION', 'config/celery/broker_region'),
                    ('CELERY_BROKER_VHOST', 'config/celery/broker_vhost'),
                ],
                'policy': 'celery_config'
            },
            'queues': {
                'map': [
                    ('CELERY_QUEUE_DEFAULT', 'config/celery/queue_default'),
                    ('CELERY_QUEUE_ENROLMENT_STORAGE', 'config/celery/queue_enrolment_storage'),
                    ('CELERY_QUEUE_ENROLMENT_VALIDATION', 'config/celery/queue_enrolment_validation'),
                    ('CELERY_QUEUE_ENROLMENT', 'config/celery/queue_enrolment'),
                    ('CELERY_QUEUE_VERIFICATION', 'config/celery/queue_verification'),
                    ('CELERY_QUEUE_ALERTS', 'config/celery/queue_alerts'),
                    ('CELERY_QUEUE_REPORTING', 'config/celery/queue_reporting'),
                ],
                'policy': 'celery_queues'
            },
            'credentials': {
                'api': {
                    'map': [
                        ('CELERY_BROKER_USER', 'config/celery/credentials/api_user'),
                        ('CELERY_BROKER_PASSWORD', 'config/celery/credentials/api_password'),
                    ],
                    'policy': 'celery_credentials_api',
                },
                'lapi': {
                    'map': [
                        ('CELERY_BROKER_USER', 'config/celery/credentials/lapi_user'),
                        ('CELERY_BROKER_PASSWORD', 'config/celery/credentials/lapi_password'),
                    ],
                    'policy': 'celery_credentials_lapi',
                },
                'worker': {
                    'map': [
                        ('CELERY_BROKER_USER', 'config/celery/credentials/worker_user'),
                        ('CELERY_BROKER_PASSWORD', 'config/celery/credentials/worker_password'),
                    ],
                    'policy': 'celery_credentials_worker',
                },
                'beat': {
                    'map': [
                        ('CELERY_BROKER_USER', 'config/celery/credentials/beat_user'),
                        ('CELERY_BROKER_PASSWORD', 'config/celery/credentials/beat_password'),
                    ],
                    'policy': 'celery_credentials_beat',
                },
                'provider': {
                    'map': [
                        ('CELERY_BROKER_USER', 'config/celery/credentials/provider_user'),
                        ('CELERY_BROKER_PASSWORD', 'config/celery/credentials/provider_password'),
                    ],
                    'policy': 'celery_credentials_provider',
                },
            }
        }
    }


def get_policies():
    """
        Get the policy definitions
        :return: Policy definition for each defined policy
        :rtype: dict
    """
    return {
        'public': {
            'path': {
                '<kv_path>/data/config/public/*': {
                    'capabilities': ['read', 'list']
                },
                '<kv_path>/data/config/tesla/domain': {
                    'capabilities': ['read', 'list']
                }
            }
        },
        'vault_vle_write': {
            "path": {
                "<kv_path>/data/modules/vle_*": {
                    "capabilities": ["read", "list", "create", "update", "delete"]
                },
                "auth/<approle_path>/role/vle_*": {
                    "capabilities": ["read", "list", "create", "update", "delete", "sudo"]
                },
                '<transit_path>/keys/jwt_module_vle_*': {
                    "capabilities": ["read", "list", "create", "update", "delete", "sudo"]
                },
            }
        },
        'vault_provider_write': {
            "path": {
                "<kv_path>/data/modules/provider_*": {
                    "capabilities": ["read", "list", "create", "update", "delete"]
                },
                "<kv_path>/data/config/celery/credentials/provider_*": {
                    "capabilities": ["read", "list", "create", "update", "delete"]
                },
                "auth/<approle_path>/role/provider_*": {
                    "capabilities": ["read", "list", "create", "update", "delete", "sudo"]
                },
                '<transit_path>/keys/jwt_module_provider_*': {
                    "capabilities": ["read", "list", "create", "update", "delete", "sudo"]
                },
            }
        },
        'vault_jwt_default_create': {
            'path': {
                '<transit_path>/sign/jwt_default': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/sign/jwt_default/*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/keys/jwt_default': {
                    'capabilities': ['read']
                },
            }
        },
        'vault_jwt_default_validate': {
            'path': {
                '<transit_path>/verify/jwt_default': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/verify/jwt_default/*': {
                    'capabilities': ['create', 'update']
                }
            }
        },
        'vault_jwt_users_create': {
            'path': {
                '<transit_path>/sign/jwt_users': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/sign/jwt_users/*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/keys/jwt_users': {
                    'capabilities': ['read']
                },
            }
        },
        'vault_jwt_users_validate': {
            'path': {
                '<transit_path>/verify/jwt_users': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/verify/jwt_users/*': {
                    'capabilities': ['create', 'update']
                }
            }
        },
        'vault_jwt_instructors_create': {
            'path': {
                '<transit_path>/sign/jwt_instructors': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/sign/jwt_instructors/*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/keys/jwt_instructors': {
                    'capabilities': ['read']
                },
            }
        },
        'vault_jwt_instructors_validate': {
            'path': {
                '<transit_path>/verify/jwt_instructors': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/verify/jwt_instructors/*': {
                    'capabilities': ['create', 'update']
                }
            }
        },
        'vault_jwt_learners_create': {
            'path': {
                '<transit_path>/sign/jwt_learners': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/sign/jwt_learners/*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/keys/jwt_learners': {
                    'capabilities': ['read']
                },
            }
        },
        'vault_jwt_learners_validate': {
            'path': {
                '<transit_path>/verify/jwt_learners': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/verify/jwt_learners/*': {
                    'capabilities': ['create', 'update']
                }
            }
        },
        'vault_jwt_module_all_validate': {
            'path': {
                '<transit_path>/verify/jwt_module_*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/verify/jwt_module_*/*': {
                    'capabilities': ['create', 'update']
                }
            }
        },
        'vault_jwt_module_all_create': {
            'path': {
                '<transit_path>/sign/jwt_module_*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/sign/jwt_module_*/*': {
                    'capabilities': ['create', 'update']
                },
                '<transit_path>/keys/jwt_module_*': {
                    'capabilities': ['read']
                },
            }
        },
        'vault_policy_write': {
            'path': {
                'sys/policy/<policy_prefix>modules_vle_*': {
                    'capabilities': ['create', 'update', 'read', 'delete', 'list', 'sudo']
                },
                'sys/policy/<policy_prefix>modules_provider_*': {
                    'capabilities': ['create', 'update', 'read', 'delete', 'list', 'sudo']
                },
            }
        },
        'db_config': {
            'path': {
                '<kv_path>/data/config/db/host': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/db/port': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/db/engine': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/db/name': {
                    'capabilities': ['read']
                },
            }
        },
        'db_credentials_admin': {
            'path': {
                '<kv_path>/data/config/db/root_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/db/root_password': {
                    'capabilities': ['read']
                },
            }
        },
        'db_credentials_rw_access': {
            'path': {
                '<kv_path>/data/config/db/user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/db/password': {
                    'capabilities': ['read']
                },
            }
        },
        'django_config': {
            'path': {
                '<kv_path>/data/config/django/*': {
                    'capabilities': ['read']
                }
            }
        },
        'redis_config': {
            'path': {
                '<kv_path>/data/config/redis/host': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/redis/port': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/redis/database': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/redis/password': {
                    'capabilities': ['read']
                },
            }
        },
        'storage_config': {
            'path': {
                '<kv_path>/data/config/storage/url': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/storage/bucket_name': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/storage/region': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/storage/access_key': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/storage/secret_key': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_config': {
            'path': {
                '<kv_path>/data/config/celery/broker_protocol': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/broker_host': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/broker_port': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/broker_region': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/broker_vhost': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_queues': {
            'path': {
                '<kv_path>/data/config/celery/queue_default': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_enrolment': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_enrolment_storage': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_enrolment_validation': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_verification': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_alerts': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/queue_reporting': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_credentials_api': {
            'path': {
                '<kv_path>/data/config/celery/credentials/api_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/credentials/api_password': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_credentials_lapi': {
            'path': {
                '<kv_path>/data/config/celery/credentials/lapi_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/credentials/lapi_password': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_credentials_worker': {
            'path': {
                '<kv_path>/data/config/celery/credentials/worker_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/credentials/worker_password': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_credentials_beat': {
            'path': {
                '<kv_path>/data/config/celery/credentials/beat_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/credentials/beat_password': {
                    'capabilities': ['read']
                },
            }
        },
        'celery_credentials_provider': {
            'path': {
                '<kv_path>/data/config/celery/credentials/provider_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/celery/credentials/provider_password': {
                    'capabilities': ['read']
                },
            }
        },
        'rabbitmq_config': {
            'path': {
                '<kv_path>/data/config/rabbitmq/admin_user': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/rabbitmq/admin_password': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/rabbitmq/admin_port': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/rabbitmq/port': {
                    'capabilities': ['read']
                },
                '<kv_path>/data/config/rabbitmq/host': {
                    'capabilities': ['read']
                },
            }
        },
    }



# Generated by Django 3.2.4 on 2021-07-09 08:23

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tesla_ce', '0015_uioption_roles'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='activityinstrument',
            name='required',
        ),
        migrations.RemoveField(
            model_name='enrolmentsamplevalidation',
            name='included',
        ),
        migrations.AlterUniqueTogether(
            name='requestproviderresult',
            unique_together={('provider', 'request')},
        ),
    ]



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Course model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .base_model import BaseModel
from .instructor import Instructor
from .learner import Learner
from .vle import VLE


class Course(BaseModel):
    """ Course model. """
    vle = models.ForeignKey(VLE, null=False, blank=False,
                            on_delete=models.CASCADE,
                            help_text=_('Course VLE.'))

    code = models.CharField(max_length=250, null=False, blank=False,
                            help_text=_("Course code."))

    vle_course_id = models.CharField(max_length=250, null=False, blank=False,
                                     help_text=_("Course id on the VLE."))

    description = models.TextField(null=True, blank=True,
                                   help_text=_("Course description."))

    start = models.DateTimeField(blank=True, null=True,
                                 help_text=_("When course starts"))

    end = models.DateTimeField(blank=True, null=True,
                               help_text=_("When course ends"))

    learners = models.ManyToManyField(Learner, blank=True, help_text=_('Course learners.'))
    instructors = models.ManyToManyField(Instructor, blank=True, help_text=_('Course instructors.'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        unique_together = (('vle', 'vle_course_id'),)

    def __repr__(self):
        return "<Course(id='%r', vle_id='%r', vle_course_id='%r', code='%s')>" % (
            self.id, self.vle_id, self.vle_course_id, self.code)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to generate the configuration file for TeSLA system """
import os

from django.core.management.base import CommandError

from tesla_ce.client import Client
from ..base import TeslaConfigCommand

BASE_TESLA_CONF_PATH = '/etc/tesla'
BASE_TESLA_CONF_FILE = 'tesla-ce.cfg'


class Command(TeslaConfigCommand):
    """ Command to generate configuration file from template """
    help = 'Generates a configuration file to be used on the initial setup of the TeSLA CE system'
    requires_system_checks = []

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        parser.add_argument(
            '--override',
            action='store_true',
            help='Override existing configuration file',
        )
        parser.add_argument(
            '--local',
            action='store_true',
            help='Create the configuration file in current directory',
        )
        parser.add_argument(
            '--with-services',
            action='store_true',
            help='Include external services data',
        )
        parser.add_argument(
            '--with-moodle',
            action='store_true',
            help='Include Moodle deployment',
        )
        parser.add_argument(
            'domain',
            help='Base domain where TeSLA will be accessible',
        )

    @staticmethod
    def get_client(config_file=None, options=None):
        """
            Create an instance of the client
            :param config_file: Path to the configuration file
            :param options: Provided options to the command
            :return: Instance of client
        """
        return Client

    def check_configuration_file(self):
        """
            Check if the defined configuration file exists
        """
        # Get the file to be checked
        # config_file = self.get_configuration_file()

        # Avoid unintentionally override
        if self.conf_file is not None and not self._options['override'] and os.path.exists(self.conf_file):
            self.stdout.write(
                self.style.ERROR('File {} already exists. Use --override option.'.format(self.conf_file))
            )
            raise CommandError('Configuration file already exists')

    def custom_handle(self):
        """
            Custom actions for this command
        """
        # Generate configuration file
        domain = self._options['domain']
        deploy_external_services = self._options['with_services']
        deploy_moodle = self._options['with_moodle']
        self.client.generate_configuration(self.conf_file, domain, deploy_external_services, deploy_moodle)

        # Change permissions to the file
        os.chmod(self.conf_file, 0o600)

        self.stdout.write(self.style.SUCCESS('Configuration file created at {}'.format(self._conf_file)))



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Default Webhook client for providers implementation """
from tesla_ce.models import Provider
from .base import BaseWebhookMessage
from .. import exceptions


class DefaultProviderWebhookMessage(BaseWebhookMessage):

    def __init__(self, client, body, message_id=None, model=None):
        super().__init__(client, body, message_id, model)
        try:
            self.provider = Provider.objects.get(acronym=client.name)
        except Provider.DoesNotExist:
            raise exceptions.WebhookProviderNotFoundException('Provider with acronym {} not found'.format(client.name))

    def _is_authenticated(self, request):
        return True

    def _process_webhook(self):
        print('PROCESSING MESSAGE')





import os
import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt", "r") as fh:
    requirements_raw = fh.read()
    requirements_list = requirements_raw.split('\n')
    requirements = []
    for req in requirements_list:
        # Skip comments and optional requirements
        if not req.strip().startswith('#') and len(req.strip()) > 0 and not req.strip().startswith(
                'mysqlclient') and not req.strip().startswith('psycopg2'):
            requirements.append(req)

with open(os.path.join("src", "tesla_ce", "lib", "data", "VERSION"), "r") as fh:
    version = fh.read()
            
setuptools.setup(
    version=version,
    name="tesla-ce",
    author="Xavier Baro",
    author_email="xbaro@uoc.edu",
    description="TeSLA CE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/tesla-ce/core",
    project_urls = {
        'Documentation': 'https://tesla-ce.github.io',
        'Source': 'https://github.com/tesla-ce/core',
    },
    packages=setuptools.find_packages('src', exclude='__pycache__'),
    package_dir={'': 'src'},  # tell distutils packages are under src
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: POSIX :: Linux",
    ],
    python_requires='>=3.6',
    include_package_data=True,
    install_requires=requirements,
    extras_require={
        '': ["mysqlclient"],
        'all': ["mysqlclient", "psycopg2"],
        'mysql': ["mysqlclient"],
        'psql': ["psycopg2"]
    },
    entry_points = {
        'console_scripts': [
            'tesla_ce=tesla_ce.manage:main',
            'manage.py=tesla_ce.manage:main'
        ],
    }    
)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution User views module """
from django.db.models import Q
from django_filters import rest_framework as filters_dj
from django_filters.rest_framework import DjangoFilterBackend
from rest_framework import viewsets
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api.v2.serializers import InstitutionUserSerializer
from tesla_ce.models import InstitutionUser


def roles_values():
    """ Get available roles for filter """
    return (
        ('ADMIN', 'ADMIN'),
        ('SEND', 'SEND'),
        ('LEGAL', 'LEGAL'),
        ('DATA', 'DATA'),
        ('INSTRUCTOR', 'INSTRUCTOR'),
        ('LEARNER', 'LEARNER'),
    )


class InstitutionUserFilter(filters_dj.FilterSet):
    """ Filter implementation for Institution Users """

    roles = filters_dj.MultipleChoiceFilter(field_name="roles",
                                            label="roles",
                                            choices=roles_values,
                                            method='filter_roles')

    username = filters_dj.CharFilter(lookup_expr='icontains')
    first_name = filters_dj.CharFilter(lookup_expr='icontains')
    last_name = filters_dj.CharFilter(lookup_expr='icontains')
    email = filters_dj.CharFilter(lookup_expr='icontains')
    uid = filters_dj.CharFilter(lookup_expr='icontains')

    def filter_roles(self, queryset, name, value):

        # No filter provided
        if value is None or len(value) == 0:
            return queryset

        # Filter by provided roles
        filter_q = Q()
        if 'ADMIN' in value:
            filter_q = filter_q | Q(inst_admin=True)
        if 'SEND' in value:
            filter_q = filter_q | Q(send_admin=True)
        if 'LEGAL' in value:
            filter_q = filter_q | Q(legal_admin=True)
        if 'DATA' in value:
            filter_q = filter_q | Q(data_admin=True)
        if 'LEARNER' in value:
            filter_q = filter_q | Q(learner__isnull=False)
        if 'INSTRUCTOR' in value:
            filter_q = filter_q | Q(instructor__isnull=False)
        return queryset.filter(filter_q)

    class Meta:
        model = InstitutionUser
        fields = ['username', 'first_name', 'last_name', 'email', 'roles', 'uid']


# pylint: disable=too-many-ancestors
class InstitutionUserViewSet(NestedViewSetMixin, viewsets.ModelViewSet):
    """
    API endpoint that allow institution users to be managed.
    """
    model = InstitutionUser
    serializer_class = InstitutionUserSerializer
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_class = InstitutionUserFilter
    search_fields = ['username', 'first_name', 'last_name', 'email', 'uid']

    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(InstitutionUser.objects)
        return queryset.all().order_by('id')



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""UI Option serialize module."""
from rest_framework import serializers

from tesla_ce.models import UIOption
from tesla_ce.models.user import get_roles


class UIOptionSerializer(serializers.ModelSerializer):
    """UI Option serialize model module."""

    id = serializers.ReadOnlyField()

    class Meta:
        model = UIOption
        fields = ['id', 'route', 'enabled', 'roles']

    def validate_route(self, value):
        """
            Check that provided route is valid
            :param value: New UI Option route
            :return: Route
        """
        if value is None:
            raise serializers.ValidationError(detail="Route cannot be empty")
        qs = UIOption.objects.filter(route=value)
        if self.instance is not None:
            qs = qs.exclude(id=self.instance.id)
        num_routes = qs.count()
        if num_routes > 0:
            raise serializers.ValidationError(detail="This route already exists")
        return value

    def validate_roles(self, value):
        """
            Check that provided roles are valid
            :param value: New UI Option route
            :return: Route
        """
        if value is None or len(value.strip()) == 0:
            return None

        valid_roles = get_roles()
        list_roles = value.split(',')
        roles = []
        for role in list_roles:
            if role.upper() not in valid_roles:
                raise serializers.ValidationError(detail="Role {} is not valid".format(role))
            roles.append(role.upper())
        return ','.join(roles)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
""" Configuration file generation test module """
import mock
import os
import tempfile
import pytest

from io import StringIO

from django.core.management import call_command
from django.core.management.base import CommandError


def test_generate_configuration(tesla_ce_system, config_mode_settings):

    from tesla_ce.lib.config import ConfigManager

    assert tesla_ce_system is not None

    out = StringIO()
    err = StringIO()

    with mock.patch('tesla_ce.management.base.TeslaCommand.get_client', return_value=tesla_ce_system):
        with tempfile.TemporaryDirectory() as tmp_dir:
            # Generate the configuration file
            conf_file = os.path.join(tmp_dir, 'test_gen_config.cfg')
            with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
                # Generate the configuration file
                domain = 'test_domain'
                call_command(
                    'generate_config',
                    domain,
                    stdout=out,
                    stderr=err,
                    with_services=False,
                    local=True
                )

            # Read the configuration
            conf = ConfigManager(load_config=False)
            conf.load_file(conf_file)

            # Check expected values
            assert conf.config.get('TESLA_DOMAIN') == 'test_domain'
            assert conf.config.get('DJANGO_SECRET_KEY') is not None
            assert 'test_domain' in conf.config.get('DJANGO_ALLOWED_HOSTS')

            # Check Moodle
            assert not conf.config.get('MOODLE_DEPLOY')
            assert conf.config.get('MOODLE_ADMIN_PASSWORD') is None
            assert conf.config.get('MOODLE_DB_HOST') is None
            assert conf.config.get('MOODLE_DB_PASSWORD') is None


def test_generate_configuration_with_services(tesla_ce_system, config_mode_settings):
    from tesla_ce.lib.config import ConfigManager

    assert tesla_ce_system is not None

    out = StringIO()
    err = StringIO()

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Generate the configuration file
        conf_file = os.path.join(tmp_dir, 'test_gen_config2.cfg')
        with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
            # Generate the configuration file
            domain = 'test_domain2'
            call_command(
                'generate_config',
                domain,
                stdout=out,
                stderr=err,
                with_services=True,
                local=True
            )

        # Read the configuration
        conf = ConfigManager(load_config=False)
        conf.load_file(conf_file)

        # Check expected values
        assert conf.config.get('TESLA_DOMAIN') == 'test_domain2'
        assert conf.config.get('DJANGO_SECRET_KEY') is not None
        assert 'test_domain2' in conf.config.get('DJANGO_ALLOWED_HOSTS')

        # Check services
        assert conf.config.get('DB_USER', None) is not None
        assert conf.config.get('DB_PASSWORD', None) is not None
        assert conf.config.get('VAULT_DB_HOST') is not None
        assert conf.config.get('VAULT_DB_USER') is not None
        assert conf.config.get('RABBITMQ_ERLANG_COOKIE') is not None
        assert conf.config.get('RABBITMQ_ADMIN_USER') is not None
        assert conf.config.get('RABBITMQ_ADMIN_PASSWORD') is not None
        assert conf.config.get('STORAGE_ACCESS_KEY') is not None
        assert conf.config.get('STORAGE_SECRET_KEY') is not None

        # Check Moodle
        assert not conf.config.get('MOODLE_DEPLOY')
        assert conf.config.get('MOODLE_ADMIN_PASSWORD') is None
        assert conf.config.get('MOODLE_DB_HOST') is None
        assert conf.config.get('MOODLE_DB_PASSWORD') is None


def test_generate_configuration_with_moodle(tesla_ce_system, config_mode_settings):
    from tesla_ce.lib.config import ConfigManager

    assert tesla_ce_system is not None

    out = StringIO()
    err = StringIO()

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Generate the configuration file
        conf_file = os.path.join(tmp_dir, 'test_gen_config3.cfg')
        with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
            # Generate the configuration file
            domain = 'test_domain3'
            call_command(
                'generate_config',
                domain,
                stdout=out,
                stderr=err,
                with_services=True,
                with_moodle=True,
                local=True
            )

        # Read the configuration
        conf = ConfigManager(load_config=False)
        conf.load_file(conf_file)

        # Check expected values
        assert conf.config.get('TESLA_DOMAIN') == 'test_domain3'
        assert conf.config.get('DJANGO_SECRET_KEY') is not None
        assert 'test_domain3' in conf.config.get('DJANGO_ALLOWED_HOSTS')

        # Check Moodle
        assert conf.config.get('MOODLE_DEPLOY')
        assert conf.config.get('MOODLE_ADMIN_PASSWORD') is not None
        assert conf.config.get('MOODLE_DB_HOST') is not None
        assert conf.config.get('MOODLE_DB_PASSWORD') is not None


def test_generate_configuration_override(tesla_ce_system, config_mode_settings):
    from tesla_ce.lib.config import ConfigManager

    assert tesla_ce_system is not None

    out = StringIO()
    err = StringIO()

    with tempfile.TemporaryDirectory() as tmp_dir:
        # Generate the configuration file
        conf_file = os.path.join(tmp_dir, 'test_gen_config4.cfg')
        with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
            # Generate the configuration file
            domain = 'test_domain4'
            call_command(
                'generate_config',
                domain,
                stdout=out,
                stderr=err,
                local=True
            )

        # Read the configuration
        conf = ConfigManager(load_config=False)
        conf.load_file(conf_file)

        # Check expected values
        assert conf.config.get('TESLA_DOMAIN') == 'test_domain4'
        assert conf.config.get('DJANGO_SECRET_KEY') is not None
        assert 'test_domain4' in conf.config.get('DJANGO_ALLOWED_HOSTS')

        with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
            # Generate again the configuration file without the override command
            domain = 'test_domain5'
            try:
                call_command(
                    'generate_config',
                    domain,
                    stdout=out,
                    stderr=err,
                    local=True
                )
                pytest.fail('Configuration file overrided')
            except CommandError:
                assert 'Use --override option' in out.getvalue()

        with mock.patch('tesla_ce.management.base.TeslaConfigCommand.get_config_file', return_value=conf_file):
            # Generate the configuration file
            domain = 'test_domain5'
            call_command(
                'generate_config',
                domain,
                stdout=out,
                stderr=err,
                override=True,
                local=True
            )

        # Read the configuration
        conf = ConfigManager(load_config=False)
        conf.load_file(conf_file)

        # Check expected values
        assert conf.config.get('TESLA_DOMAIN') == 'test_domain5'
        assert conf.config.get('DJANGO_SECRET_KEY') is not None
        assert 'test_domain5' in conf.config.get('DJANGO_ALLOWED_HOSTS')



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Test module for approle authentication """
import pytest


@pytest.mark.django_db
def test_api_authentication_approle(rest_api_client, api_client, institution_course_test_case, providers):

    # Generate a approle for a VLE
    vle_creds = api_client.vault.register_vle(institution_course_test_case['vle'])

    # Generate a approle for a Provider
    provider_creds = api_client.vault.register_provider(providers['fr'])

    # Authenticate using vle credentials
    resp_vle = rest_api_client.post('/api/v2/auth/approle', data=vle_creds)

    assert resp_vle is not None
    assert resp_vle.status_code == 200
    assert resp_vle.data['pk'] == institution_course_test_case['vle'].id
    assert resp_vle.data['module'].startswith('vle_')
    assert resp_vle.data['type'] == 'vle'

    validation_jwt_vle = api_client.validate_token(resp_vle.data['token']['access_token'])
    assert validation_jwt_vle is not None
    assert isinstance(validation_jwt_vle, dict)
    assert validation_jwt_vle['group'].startswith('module_')
    assert validation_jwt_vle['pk'] == institution_course_test_case['vle'].id
    assert validation_jwt_vle['type'] == 'vle'

    # Authenticate using provider credentials
    resp_provider = rest_api_client.post('/api/v2/auth/approle', data=provider_creds)

    assert resp_provider is not None
    assert resp_provider.status_code == 200
    assert 'pk' in resp_provider.data
    assert resp_provider.data['module'].startswith('provider_')
    assert resp_provider.data['type'] == 'provider'

    validation_jwt_provider = api_client.validate_token(resp_provider.data['token']['access_token'])
    assert validation_jwt_provider is not None
    assert isinstance(validation_jwt_provider, dict)
    assert validation_jwt_provider['group'].startswith('module_')
    assert 'pk' in validation_jwt_provider
    assert validation_jwt_provider['type'] == 'provider'



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" TeSLA CE Global Administrator actions for Use Case tests """
import json

from tests import auth_utils
from tests.conftest import get_random_string
from tests.utils import get_provider_desc_file


def api_register_providers(global_admin):
    """
        A global administrator register the providers
        :param global_admin: Global admin object
        :return: Registered providers
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(global_admin)

    # Get the list of instruments
    instruments = []
    stop = False
    while not stop:
        list_instruments_resp = client.get('/api/v2/admin/instrument/?offset={}'.format(len(instruments)))
        assert list_instruments_resp.status_code == 200
        instruments += list_instruments_resp.data['results']

        if len(instruments) == list_instruments_resp.data['count']:
            stop = True

    # Get the instruments we want to register
    fr_inst = None
    ks_inst = None
    plag_inst = None

    # Remove any existing provider
    for instrument in instruments:
        if instrument['acronym'] == 'fr':
            fr_inst = instrument['id']
        if instrument['acronym'] == 'ks':
            ks_inst = instrument['id']
        if instrument['acronym'] == 'plag':
            plag_inst = instrument['id']
        stop = False
        while not stop:
            list_inst_providers_resp = client.get('/api/v2/admin/instrument/{}/provider/'.format(instrument['id']))
            assert list_inst_providers_resp.status_code == 200
            if list_inst_providers_resp.data['count'] > 0:
                for inst_prov in list_inst_providers_resp.data['results']:
                    del_providers_resp = client.delete(
                        '/api/v2/admin/instrument/{}/provider/{}/'.format(instrument['id'], inst_prov['id'])
                    )
                    assert del_providers_resp.status_code == 204
            else:
                stop = True

    providers = {}

    # Register a FR provider
    fr_desc = json.load(open(get_provider_desc_file('fr_tfr'), 'r'))
    fr_desc['enabled'] = True
    fr_desc['validation_active'] = True
    if 'instrument' in fr_desc:
        del fr_desc['instrument']
    fr_prov_register_resp = client.post('/api/v2/admin/instrument/{}/provider/'.format(fr_inst),
                                        data=fr_desc,
                                        format='json')
    assert fr_prov_register_resp.status_code == 201
    providers['fr'] = fr_prov_register_resp.data
    providers['fr']['deferred'] = False

    # Register a FR provider that use deferred analysis
    fr_desc2 = json.load(open(get_provider_desc_file('fr_amazon'), 'r'))
    fr_desc2['enabled'] = True
    fr_desc2['validation_active'] = True
    if 'instrument' in fr_desc2:
        del fr_desc2['instrument']
    fr_prov2_register_resp = client.post('/api/v2/admin/instrument/{}/provider/'.format(fr_inst),
                                         data=fr_desc2,
                                         format='json')
    assert fr_prov2_register_resp.status_code == 201
    providers['fr_def'] = fr_prov2_register_resp.data
    providers['fr_def']['deferred'] = True

    # Register a KS provider
    ks_desc = json.load(open(get_provider_desc_file('ks_tks'), 'r'))
    ks_desc['enabled'] = True
    ks_desc['validation_active'] = True
    if 'instrument' in ks_desc:
        del ks_desc['instrument']
    ks_prov_register_resp = client.post('/api/v2/admin/instrument/{}/provider/'.format(ks_inst),
                                        data=ks_desc,
                                        format='json')
    assert ks_prov_register_resp.status_code == 201
    providers['ks'] = ks_prov_register_resp.data
    providers['ks']['deferred'] = False

    # Register a Plagiarism provider
    pt_desc = json.load(open(get_provider_desc_file('pt_tpt'), 'r'))
    pt_desc['enabled'] = True
    if 'instrument' in pt_desc:
        del pt_desc['instrument']
    pt_prov_register_resp = client.post('/api/v2/admin/instrument/{}/provider/'.format(plag_inst),
                                        data=pt_desc,
                                        format='json')
    assert pt_prov_register_resp.status_code == 201
    providers['plag'] = pt_prov_register_resp.data
    providers['plag']['deferred'] = False

    # Register a Plagiarism provider with deferred processing
    pt_desc2 = json.load(open(get_provider_desc_file('pt_turkund'), 'r'))
    pt_desc2['enabled'] = True
    if 'instrument' in pt_desc2:
        del pt_desc2['instrument']
    pt2_prov_register_resp = client.post('/api/v2/admin/instrument/{}/provider/'.format(plag_inst),
                                         data=pt_desc2,
                                         format='json')
    assert pt2_prov_register_resp.status_code == 201
    providers['plag_def'] = pt2_prov_register_resp.data
    providers['plag_def']['deferred'] = True

    # Enable the instruments
    fr_inst_enable_resp = client.patch('/api/v2/admin/instrument/{}/'.format(fr_inst),
                                       data={'enabled': True},
                                       format='json')
    assert fr_inst_enable_resp.status_code == 200
    ks_inst_enable_resp = client.patch('/api/v2/admin/instrument/{}/'.format(ks_inst),
                                       data={'enabled': True},
                                       format='json')
    assert ks_inst_enable_resp.status_code == 200
    pt_inst_enable_resp = client.patch('/api/v2/admin/instrument/{}/'.format(plag_inst),
                                       data={'enabled': True},
                                       format='json')
    assert pt_inst_enable_resp.status_code == 200

    return providers


def api_create_institution(global_admin, external_ic=False):
    """
        A global admin creates a new institution
        :param global_admin: Global admin object
        :param external_ic: True if the IC is managed externally, and assumed accepted, False otherwise
        :return: New created institution
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(global_admin)

    # Create a new institution
    institution_data = {
        'name': "PyTest Test institution {}".format(get_random_string(5)),
        'acronym': get_random_string(10),
        'external_ic': external_ic
    }
    inst_create_resp = client.post('/api/v2/admin/institution/', data=institution_data, format='json')
    assert inst_create_resp.status_code == 201

    # Get institution object
    institution = inst_create_resp.data

    return institution


def api_create_institution_admin(global_admin, institution):
    """
        A global administrator creates a new user for the institution and assign administration rights
        :param global_admin: Global admin object
        :param institution: Institution object
        :return: New created institution
    """
    # Set global admin user.
    client = auth_utils.client_with_user_obj(global_admin)

    # Create a new user
    user_name = get_random_string(10)
    password = get_random_string(10)
    email = '{}@tesla-ce.eu'.format(user_name)
    user_data = {
        'username': user_name,
        'uid': user_name,
        'email': email,
        'first_name': user_name[:5],
        'last_name': user_name[5:],
        'institution_id': institution['id'],
        'login_allowed': True,
        'is_active': True,
        'is_staff': False,
        'password': password,
        'password2': password,
        'inst_admin': True
    }
    user_create_resp = client.post('/api/v2/admin/user/', data=user_data, format='json')
    assert user_create_resp.status_code == 201

    # Get institution admin object
    admin_user = user_create_resp.data
    assert 'ADMIN' in admin_user['roles']
    assert admin_user['institution']['id'] == institution['id']

    # Return credentials
    return {
        'email': email,
        'password': password
    }



#  Copyright (c) 2020 Roger Muñoz
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Report Activity model module."""

from django.db import models
from django.utils.translation import gettext_lazy as _

from .assessment_session import AssessmentSession
from .base_model import BaseModel
from .report_activity import ReportActivity
from .report_activity import REPORT_ALERT_LEVEL


class ReportActivitySession(BaseModel):
    """ Report for an Activity Assessment Session model. """

    report = models.ForeignKey(ReportActivity, null=False, blank=False, on_delete=models.CASCADE,
                               help_text=_('Related Activity Report.'))

    session = models.ForeignKey(AssessmentSession, null=False, blank=False, on_delete=models.CASCADE, unique=True,
                                help_text=_('Assessment Session related to this report.'))

    identity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                              help_text=_('Alert level for learner identity.'))
    content_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                             help_text=_('Alert level for content authorship.'))
    integrity_level = models.SmallIntegerField(choices=REPORT_ALERT_LEVEL, null=False, default=0,
                                               help_text=_('Alert level for system integrity.'))

    total_requests = models.IntegerField(null=False, default=0,
                                         help_text=_('Number of requests on this session.'))
    pending_requests = models.IntegerField(null=False, default=0,
                                           help_text=_('Number of pending requests on this session.'))
    valid_requests = models.IntegerField(null=False, default=0,
                                         help_text=_('Number of valid requests on this session.'))
    processed_requests = models.IntegerField(null=False, default=0,
                                             help_text=_('Number of processed requests on this session.'))

    data = models.FileField(max_length=250, null=True, blank=False, default=None,
                            help_text=_("Path to the content of this report."))

    closed_at = models.DateTimeField(null=True, default=None, help_text=_('When the session report was closed'))

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    def __repr__(self):
        return "<ReportActivitySession(id='%r', session_id='%r', learner_id='%r' activity_id='%r')>" % (
            self.id, self.session.id, self.session.learner.learner_id, self.session.activity_id)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Assessment session data model module."""
from django.db import models
from django.utils.translation import gettext_lazy as _

from .assessment_session import AssessmentSession
from .base_model import BaseModel


class AssessmentSessionData(BaseModel):
    """ Assessment session data model """
    session = models.OneToOneField(AssessmentSession, null=False,
                                   on_delete=models.CASCADE, help_text=_('Related assessment session'))

    connector = models.FileField(help_text=_('Connector JS file for this session'))

    data = models.FileField(help_text=_('Data for this session'))

    def __repr__(self):
        return "<AssessmentSessionData(session_id='%r')>" % (self.session_id)

    def delete(self, using=None, keep_parents=False):
        # TODO: Remove data from storage
        return super().delete(using, keep_parents)



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango command to generate deployment scripts for a VLE """
from ..base import TeslaDeployCommand
from tesla_ce import models
from tesla_ce import tasks


class Command(TeslaDeployCommand):
    """ Command to resend all pending enrolment samples """
    help = 'Resend all pending enrolment samples'
    requires_system_checks = '__all__'

    def add_arguments(self, parser):
        """
            Define custom arguments for this command

            :param parser: Input command parser instance
        """
        # Set default arguments
        super().add_arguments(parser)

        parser.add_argument(
            '--institution-id',
            help='Institution Id',
            default=None,
        )

        parser.add_argument(
            '--user-id',
            help='User Id',
            default=None,
        )

        parser.add_argument(
            '--include-processed',
            action='store_true',
            help='Resend also processed samples',
        )

    def custom_handle(self):
        """
            Custom actions for this command
        """
        # Get all samples
        samples = models.EnrolmentSample.objects

        # Filter by institution
        if self._options['institution_id'] is not None:
            samples = samples.filter(learner__institution_id=self._options['institution_id'])

        # Filter by user
        if self._options['user_id'] is not None:
            samples = samples.filter(learner_id=self._options['user_id'])

        # Filter processed
        if not self._options['include_processed']:
            samples = samples.filter(status__in=[0, 3, 4])  # Valid and error samples are not included

        samples_count = 0
        error_count = 0
        for sample in samples.all():
            try:
                tasks.requests.enrolment.validate_request(sample.learner_id, sample.id)
            except Exception:
                error_count += 1
            samples_count += 1

        if error_count == 0:
            self.stdout.write(self.style.SUCCESS('{} samples sent'.format(samples_count)))
        else:
            self.stdout.write(self.style.ERROR('{} samples sent with {} errors'.format(samples_count, error_count)))



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Learner Verification Requests Serializers """
from rest_framework import serializers


class VerificationMetadataSerializer(serializers.Serializer):


    filename = serializers.CharField(required=False, allow_null=True, allow_blank=False, default=None)
    mimetype = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    context = serializers.JSONField(required=False, allow_null=True, default=None)

    class Meta:
        fields = ["filename", "mimetype", "context"]


class VerificationSampleSerializer(serializers.Serializer):
    """ Body serializer for verification request."""

    learner_id = serializers.UUIDField(required=True, allow_null=False)
    course_id = serializers.IntegerField(required=True, allow_null=False)
    activity_id = serializers.IntegerField(required=True, allow_null=False)
    session_id = serializers.IntegerField(allow_null=True, default=None)
    data = serializers.CharField(required=True, allow_null=False, allow_blank=False)
    instruments = serializers.ListField(required=True, allow_null=False, allow_empty=False,
                                        child=serializers.IntegerField(allow_null=False))
    metadata = VerificationMetadataSerializer(required=True, allow_null=False)

    class Meta:
        fields = ["learner_id", "course_id", "activity_id", "session_id", "data", "instruments", "metadata"]



#  Copyright (c) 2020 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" Institution Learner views module """
from django.utils import timezone

from django_filters.rest_framework import DjangoFilterBackend

from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.filters import OrderingFilter
from rest_framework.filters import SearchFilter
from rest_framework.views import Response
from rest_framework.views import status
from rest_framework.exceptions import PermissionDenied
from rest_framework_extensions.mixins import DetailSerializerMixin
from rest_framework_extensions.mixins import NestedViewSetMixin

from tesla_ce.apps.api import permissions
from tesla_ce.apps.api.v2.serializers import InstitutionLearnerDetailSerializer
from tesla_ce.apps.api.v2.serializers import InstitutionLearnerSerializer
from tesla_ce.apps.api.v2.serializers import InstitutionLearnerICBodySerializer

from tesla_ce.models import InformedConsent
from tesla_ce.models import Learner
from tesla_ce.models.user import get_institution_user
from tesla_ce.models.user import is_global_admin


def is_newer_versions(current, new):
    """
        Compare if a version is newer than the current one
        :param current: Current version in standard format x.y.z
        :type current: str
        :param new: New version in standard format x.y.z
        :type new: str
        :return: True if the new version is newer than the current. False otherwise.
    """
    current_version = [int(ver_part) for ver_part in current.split('.')]
    new_version = [int(ver_part) for ver_part in new.split('.')]
    if new_version[0] > current_version[0] or\
            (new_version[0] == current_version[0] and
            (
                    new_version[1] > current_version[1] or
                    (new_version[1] == current_version[1] and new_version[2] > current_version[2])
            )):
        return True
    return False


# pylint: disable=too-many-ancestors
class InstitutionLearnerViewSet(NestedViewSetMixin, DetailSerializerMixin, viewsets.ModelViewSet):
    """
    API endpoint that allows learners to be viewed or edited.
    """
    model = Learner
    serializer_class = InstitutionLearnerSerializer
    serializer_detail_class = InstitutionLearnerDetailSerializer
    permission_classes = [
        permissions.GlobalAdminReadOnlyPermission |
        permissions.InstitutionAdminPermission |
        permissions.InstitutionLegalAdminReadOnlyPermission |
        permissions.InstitutionDataAdminReadOnlyPermission |
        permissions.InstitutionLearnerReadOnlyPermission
    ]
    filter_backends = [SearchFilter, DjangoFilterBackend, OrderingFilter]
    filterset_fields = ['uid', 'email', 'first_name', 'last_name']
    search_fields = ['uid', 'email', 'first_name', 'last_name']

    def get_queryset(self):
        queryset = self.filter_queryset_by_parents_lookups(Learner.objects)
        if not is_global_admin(self.request.user):
            inst_user = get_institution_user(self.request.user)
            if not inst_user.inst_admin and not inst_user.legal_admin and not inst_user.data_admin:
                queryset = queryset.filter(id=inst_user.id)
        return queryset.all().order_by('id')

    @action(detail=True, methods=['POST', 'DELETE'], permission_classes=[permissions.InstitutionLearnerPermission])
    def ic(self, request, *args, **kwargs):
        """
            Manage learner informed consent
        """
        learner = Learner.objects.get(
            pk=kwargs['pk'],
            institution_id=kwargs['parent_lookup_institution_id']
        )

        if learner.id != request.user.id:
            raise PermissionDenied('Only learner can manage his/her IC')

        if request.method == 'POST':
            # Accept informed consent
            serializer = InstitutionLearnerICBodySerializer(data=request.data)
            if serializer.is_valid():
                try:
                    consent = InformedConsent.objects.get(institution=learner.institution,
                                                          version=serializer.validated_data['version'])
                    if learner.consent is not None and \
                            learner.consent_rejected is None and \
                            not is_newer_versions(learner.consent.version, consent.version):
                        return Response('Cannot downgrade consent', status=status.HTTP_400_BAD_REQUEST)

                except InformedConsent.DoesNotExist:
                    return Response('Informed Consent not found', status=status.HTTP_404_NOT_FOUND)

                # Update the informed consent of the learner
                learner.consent = consent
                learner.consent_accepted = timezone.now()
                learner.consent_rejected = None
                learner.save()
            else:
                return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

        elif request.method == 'DELETE':
            # Reject the current informed consent for this learner
            if learner.consent is not None and learner.consent_rejected is None:
                learner.consent_rejected = timezone.now()
                learner.save()
            else:
                return Response('No Informed Consent to reject', status=status.HTTP_404_NOT_FOUND)

        serializer = self.get_serializer(learner)
        return Response(serializer.data)

    @action(detail=True, methods=['GET'])
    def enrolment(self, request, *args, **kwargs):
        """
            Get learner enrolment status
        """
        learner = Learner.objects.get(
            pk=kwargs['pk'],
            institution_id=kwargs['parent_lookup_institution_id']
        )

        if 'activity_id' in request.query_params:
            return Response(learner.missing_enrolments(int(request.query_params['activity_id'])))

        return Response(learner.enrolment_status)



#  Copyright (c) 2021 Xavier Baró
#
#      This program is free software: you can redistribute it and/or modify
#      it under the terms of the GNU Affero General Public License as
#      published by the Free Software Foundation, either version 3 of the
#      License, or (at your option) any later version.
#
#      This program is distributed in the hope that it will be useful,
#      but WITHOUT ANY WARRANTY; without even the implied warranty of
#      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#      GNU Affero General Public License for more details.
#
#      You should have received a copy of the GNU Affero General Public License
#      along with this program.  If not, see <https://www.gnu.org/licenses/>.
""" DJango Configuration tests module """
import os
import mock
from tesla_ce.settings import Production, Dev
from tesla_ce.lib import ConfigManager
from tesla_ce.lib.modules import get_modules


def check_config(settings_cls, module, debug=False, worker=False):
    """
        Check configuration class values
        :param settings_cls: The configuration class
        :param module: The name of the module
        :param debug: If configuration should have debug enabled
        :param worker: If configuration corresponds to a worker
    """
    assert settings_cls.DEBUG == debug
    assert settings_cls.TESLA_CONFIG is not None
    assert len(settings_cls.TESLA_MODULES) == 1
    assert module in settings_cls.TESLA_MODULES
    assert len(settings_cls.TESLA_CONFIG.enabled_modules) == 1
    assert module in settings_cls.TESLA_CONFIG.enabled_modules
    assert len(settings_cls.TESLA_CONFIG.modules) == 1
    assert module in settings_cls.TESLA_CONFIG.modules
    assert settings_cls.TESLA_CONFIG.modules[module]['module'] == module

    if worker:
        assert settings_cls.CELERY_BROKER_URL is not None
        # assert settings_cls.AWS_ACCESS_KEY_ID is not None
        # assert settings_cls.AWS_SECRET_ACCESS_KEY is not None


def get_available_modules():
    """
        Get the list of modules and workers defined in TeSLA CE
        :return: Tuple with modules and workers names
    """
    modules = []
    workers = []
    available_modules = get_modules()
    for module in available_modules:
        modules.append(module)
        if 'type' in available_modules[module]['deployment'] and \
                available_modules[module]['deployment']['type'] == 'worker':
            workers.append(module)

    return modules, workers


def test_production_config(admin_client):
    modules, workers = get_available_modules()
    for module in modules:
        creds = admin_client.vault.get_module_credentials(module)
        with mock.patch.dict(os.environ, {
            "VAULT_URL": admin_client.config.config.get('VAULT_URL'),
            "VAULT_ROLE_ID": creds['role_id'],
            "VAULT_SECRET_ID": creds['secret_id']
        }, clear=True):
            Production.TESLA_CONFIG = ConfigManager()
            Production.pre_setup()
            Production.post_setup()
            is_worker = module in workers
            check_config(Production, module, debug=False, worker=is_worker)


def test_dev_config(admin_client):
    modules, workers = get_available_modules()
    for module in modules:
        creds = admin_client.vault.get_module_credentials(module)
        with mock.patch.dict(os.environ, {
            "VAULT_URL": admin_client.config.config.get('VAULT_URL'),
            "VAULT_ROLE_ID": creds['role_id'],
            "VAULT_SECRET_ID": creds['secret_id']
        }, clear=True):
            Dev.TESLA_CONFIG = ConfigManager()
            Dev.pre_setup()
            Dev.post_setup()
            is_worker = module in workers
            check_config(Dev, module, debug=True, worker=is_worker)


def test_dev_config_multiple(admin_client):
    token = admin_client.config.config.get('VAULT_TOKEN')
    with mock.patch.dict(os.environ, {
        "VAULT_URL": admin_client.config.config.get('VAULT_URL'),
        "VAULT_TOKEN": token,
        "TESLA_RUN_AS_MODULE": 'api,lapi',
        "RABBITMQ_ADMIN_USER": admin_client.config.config.get('RABBITMQ_ADMIN_USER'),
        "RABBITMQ_ADMIN_PASSWORD": admin_client.config.config.get('RABBITMQ_ADMIN_PASSWORD')
    }):
        Dev.TESLA_CONFIG = ConfigManager()
        Dev.pre_setup()
        Dev.post_setup()

        assert len(Dev.TESLA_MODULES) == 2
        assert 'api' in Dev.TESLA_MODULES
        assert 'lapi' in Dev.TESLA_MODULES
